 
/* (c) Copyright 1993 by Jeff Thorne.  This program can be copied as long as
   this message is not deleted and the program is distributed free of charge */
 
#include <math.h>
#include <stdio.h>
#include "common.h"
#ifdef PROT
#include "prottg.h"
#else
#include "dna.h"
#endif

extern double bfrq[NUMLETTERS+1]; 

void 
unrelated(seq,pars,gtol)
int **seq;
double *pars,gtol;
{

   int ktrack;
   double oldrat,newrat, oldtgap,newtgap,oldhood,newhood;
   double *seqstat, **dumstat1,**dumstat2;
   double thehood;
   extern double **dmat();
   extern double *dvec();
   extern void freedmat();
   extern void freedvec();
   extern double gsearch();
   double unreldens();
   
   dumstat1=dmat(0,0);
   dumstat2=dmat(0,0);
   seqstat=dvec(1); 

   seqstat[0]= (double) seq[0][0];
   seqstat[1]= (double) seq[1][0];

   oldhood= -1.0*BIGNUM;
   newhood= -0.5*BIGNUM;

   oldrat= (seqstat[0]+seqstat[1])/(seqstat[0]+ seqstat[1] + 2.0);
   oldtgap= oldrat;

   while(newhood-oldhood > gtol)
   {
     oldhood=newhood;
     newrat=gsearch((oldrat+0.001)/2.0, 0.9*oldrat+0.005, (1.0+oldrat)/2.0,
             unreldens,1,oldtgap,dumstat1,dumstat2,seqstat,gtol/10.0);
     newtgap=gsearch((oldtgap+0.001)/2.0, 0.9*oldtgap+0.005, (1.0+oldtgap)/2.0,
             unreldens,2,newrat,dumstat1,dumstat2,seqstat, gtol/10.0);
     newhood=unreldens(newrat, 1, newtgap,dumstat1,dumstat2,seqstat);
     /* 
        printf("newhood= %10.8E oldhood= %10.8E newrat= %6.4f newtgap= %6.4f\n",
            newhood, oldhood, newrat, newtgap); 
     */
     oldrat=newrat;
     oldtgap=newtgap;
   }

   thehood=newhood;
   for (ktrack=1;ktrack<=seq[0][0];ktrack++)
   {
      thehood += log(bfrq[seq[0][ktrack]]);
   }

   for (ktrack=1;ktrack<=seq[1][0];ktrack++)
   {
      thehood += log(bfrq[seq[1][ktrack]]);
   }

   printf("Unrelated Log-Lhood= %9.7E (",thehood);
   printf("newrat= %8.6E,", newrat);
   printf("\n newtgap= %8.6E)\n", newtgap);

   freedvec(seqstat,1);
   freedmat(dumstat1,0,0);
   freedmat(dumstat2,0,0);
}

/* calculate an approximate likelihood for the sequences under the 
   assumption that they are unrelated.   Just an interesting number, 
   avoid statistical interpretation until further notice */ 
double 
unreldens(thispar,parflag,othpar,dumstat1,dumstat2,seqstat)  
double thispar,othpar, **dumstat1, **dumstat2;
double *seqstat; 
int parflag;
{
   int len1,len2,ktrack1,ktrack2,tempee,twojumps;
   double lmurat,tgap;
   double thehood, tg,sf,dumtg,dumsf,micronum,magnify;
 
   len1= (int) seqstat[0];
   len2= (int) seqstat[1];

   if (parflag==1)
   {
      lmurat=thispar;
      tgap=othpar;
   }
   else
   /* parflag ==2 */
   {
      lmurat=othpar;
      tgap=thispar;
   }
   magnify= exp(2.0*JUMP);
   micronum= exp(-1.0*JUMP); 
   twojumps= JUMP+JUMP;
   tempee= JUMP;
   tg=exp(JUMP*1.0)*(1.0-tgap)*(1.0-tgap)*(1.0-lmurat)*(1.0-lmurat);
   for (ktrack1=1;ktrack1<= len1;ktrack1++)
   {
      if (ktrack1==1)
      {
         sf=lmurat*tg;
         tg= tg*0.5*tgap;
      }
      else
      {
         sf=lmurat*(sf+ tg);
         tg *= tgap;
      }
      if (sf+tg < micronum)
      {
         tempee += twojumps;
         sf *= magnify;
         tg *= magnify;
      }      
   }
   
   dumsf=sf;
   dumtg=tg;
   for (ktrack2=1;ktrack2<= len2;ktrack2++)
   {
      if (ktrack2==1)
      {
         tg = tgap*0.5*(dumtg+dumsf);
         sf = dumtg*lmurat;
      }
      else
      {
         tg = (0.5*sf+tg)*tgap;
         sf *= lmurat; 
      }
      if (sf+tg < micronum)
      {
         sf *= magnify;
         tg *= magnify;
         tempee += twojumps;
      }
   }

   if ( sf+tg != 0.0) 
      thehood= log(sf+tg)-tempee;
   else 
      thehood = -BIGNUM;
   tg=exp(JUMP*1.0)*(1.0-tgap)*(1.0-tgap)* (1.0-lmurat)*(1.0-lmurat);
   tempee=JUMP;
   
   for (ktrack2=1;ktrack2<= len2;ktrack2++)
   {
      if (ktrack2==1)
      {
         sf= lmurat*tg;
         tg= tgap*0.5*tg;
      }
      else
      {
         sf= lmurat*(tg+sf);
         tg *= tgap;
      }
      if (tg+sf < micronum)
      {
         tg *= magnify;
         sf *= magnify;
         tempee += twojumps;
      }
   }

   dumsf=sf;
   dumtg=tg;
   for (ktrack1=1;ktrack1<= len1;ktrack1++)
   {
      if (ktrack1==1)
      {
         tg= tgap*0.5*(dumsf+dumtg);
         sf= lmurat*(dumsf+dumtg);
      }
      else
      {
         tg = tgap*(0.5*sf+tg);
         sf *= lmurat;
      }
      if (tg+sf < micronum)
      {
         tg *= magnify;
         sf *= magnify;
         tempee += twojumps;
      }
   }
   thehood= log(tg+sf)-tempee + log(1.0+exp(thehood-log(tg+sf)+tempee));
   return(thehood);
}
