 
/* (c) Copyright 1993 by Jeff Thorne.  This program can be copied as long as
   this message is not deleted and the program is distributed free of charge */
 
#include<stdio.h>
#include <math.h>
#include "common.h"
#ifdef PROT
#include "prottg.h" 

/* needed when slow substitution param is to be estimated but ratio between
   fast and slow rates is fixed */ 
double 
fspam(thispar,parflag,othpar, fstat, sstat, othstat)
double thispar, othpar, **fstat, **sstat, *othstat;
int parflag;
{
   double fpam();
   double thedens;

   thedens= fpam(thispar,parflag,othpar,fstat, sstat, othstat);
   thedens += fpam(thispar*othpar, parflag, othpar, sstat, fstat,othstat); 
   
   return(thedens);
}
#else
#include "dna.h"

/* Returns the substitution density desired by routine gsearch.  It
   is the "interpeter" between routine gsearch and routine fsub.
   The value of parflag specifies the parameter being estimated by
   gsearch: -1 means general substitution parameter, -5 means ratio between 
   fast and slow substitution rates, -7 means within group substitution 
   parameter */
double 
fslsub(thispar,parflag,othpar,fstat,sstat,pars)
double thispar, othpar, **fstat, **sstat, *pars;
int parflag;
{
   double fsub();
   double thegen,thewig,thek,thedens;

   if (parflag== -1)
   {
      thegen=thispar;
      thewig=pars[7];
      thek=pars[5];
   }
   else
   {
      if (parflag== -5)
      {
         thek=thispar;
         thegen=pars[1];
         thewig=pars[7];
      }
      else 
      {
         if ( parflag== -7) 
         {
            thewig=thispar;
            thegen=pars[1];
            thek=pars[5];
         }
         else
         {
            printf("Trouble in routine fslsub !\n");
         }
      }
   }

   thedens=fsub(thegen,parflag,thewig,fstat,sstat, pars);
   thedens += fsub(thegen*thek, parflag, thewig*thek,sstat,fstat, pars);

   return(thedens);
}

/* called by gsearch, talks to fsub when only ratio between fast and slowly
   evolving regions is to be estmated */
double 
ktofsub(thispar,parflag,othpar,fstat,sstat,pars)
double thispar, othpar, **sstat, **fstat, *pars;
int parflag;
{
   double fsub();

   return(fsub(pars[1]*thispar,parflag,pars[7]*thispar,sstat,fstat,pars));
}

/* substitution rate: log posterior density for sequences 
   uses thispar as substitution parameter and substat as
   observed substitution or replacement counts */
double 
fsub(thispar,parflag,othpar,othsubstat,substat, othstat)
#endif

#ifdef PROT
double 
fpam(thispar,parflag,othpar,othsubstat,substat, othstat)
#endif

double thispar,othpar,**substat,**othsubstat,*othstat;
int parflag;
{
   int ktrack1,ktrack2;
   double d, **thisf;
   extern void dobd();
   extern double **dmat();
   extern void freedmat();

   thisf= dmat(NUMLETTERS,NUMLETTERS);
   d=0.0;

#ifdef PROT
   dobd(thispar,thisf);
#else
   dobd(thispar,othpar, thisf);
#endif

   /* on line below, don't need to check ktrack1==NUMLETTERS because
      amino acid frequencies are fixed */
   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {
      d += substat[ktrack1][ktrack1]*log(thisf[ktrack1][ktrack1]);
      for (ktrack2=0;ktrack2<ktrack1;ktrack2++)
      {
         d += (substat[ktrack1][ktrack2]+substat[ktrack2][ktrack1])* 
         log(thisf[ktrack1][ktrack2]);
      }
   }
   /* return log-probability */
   freedmat(thisf,NUMLETTERS,NUMLETTERS);
   return(d);
}

/* insertion rate: log posterior density, called by routine gsearch
   if parflag > 0 then  gsearch is maximizing ratio of insertion
   rate to deletion rate else it's maximizing the deletion parameter -
   the absolute value of parflag MUST be the sum of the two sequence
   lengths */
double 
fmulmu(thispar,parflag,othpar,othsubstat,substat,othstat) 
double thispar,othpar,**substat,**othsubstat,*othstat;
int parflag;
{
        double u,v,h, d; 
        double lmurat,themu;

        if (parflag>0)
        {
           lmurat=thispar;
           themu=othpar;
        }
        else
        {
           lmurat=othpar;
           themu=thispar;
        }
        u=exp(-themu);
	h = (1.0 - pow(u, 1.0-lmurat))/(1.0 - lmurat*pow(u, 1.0-lmurat));
        v= 1.0-u-h;
        if(v<= 0.0)
        {
           /* the exp(x) function may be inaccurate for small x */
           v=0.0;
           if(othstat[3] > SMALLNUM)
           {
              printf("trouble in routine fmulmu()?\n");
              printf("possible non-negligible inaccuracy\n");
              printf(" due to inaccurate exp() ");
              printf("function provided by computer environment.\n");
              printf("DEBUG:othstat[3]= %10.8E u= %10.8E h=%10.8E\n",
                      othstat[3],u,h);
           }
        } 
        else 
        {
           v = othstat[3]*log(v);
        }
 
	d =  -1.0*(othstat[1]+othstat[2])*themu + othstat[4]*log(h)
		+ v + (othstat[1]+ othstat[2] +
             othstat[3]+1.0)*log(1.0-lmurat*h) +  
         log(1.0-lmurat)+(othstat[1]+othstat[2]+othstat[3]+othstat[4])*
         log(lmurat);

	/* return log-probability */
	return(d);
}

