 
/* (c) Copyright 1993 by Jeff Thorne.  This program can be copied as long as
   this message is not deleted and the program is distributed free of charge */
 
#include<stdio.h>
#ifdef PROT
#include "prottg.h" 
#else
#include "dna.h"
#endif
#include "common.h"
 
void 
domin(seq,pars,estim,mask)  /* controls EM routine */
int **seq, *estim;
double *pars, **mask;
{
   double **holdtg,**hold[7];
   int ktrack,**holdee;
   int numits;  /* number of iterations of EM routine */
   double gtol; /* How accurate should specific parameter estimates be for a
                   given set of statistics */
   double newhood,oldhood;
   double **f,**fastf; /* slow substitution probs., fast substitution probs.
                          Both multiplied by probability of initial a.a. */
   double **fstat,**sstat; /* EM statistics for fast and slow substitutions */
   double *othstat; /* the other "EM" statistics */
   int *nestim; /* contains the info. about whether user wants parameter
                   estimated plus whether parameter is in allowable range */
   int numsamps;

#ifdef PROT
   extern double fpam(); /* function for replacement likelihoods */
   extern double fspam(); /* function needed when ratio between fast and slow
                             rates is fixed, but slow amount not fixed */
#else
   extern void estdnasub(); /* estimates substitution parameters for DNA */
#endif
   extern double fmulmu(); /* function for insertion-deletion likelihood */
   extern double dubsum();
   extern double gsearch(); /* golden section search */
   extern void mulmusearch(); /* joint estimates of mu and r */
   extern void expect(); /* E part of EM */
   extern void expectps();
   extern double **dmat();
   extern int **imat();
   extern double *dvec();
   extern int *ivec();
   extern void freeivec();
   extern void freedvec();
   extern void freeimat();
   extern void freedmat();

   numsamps =seq[0][0]/SAMPFREQ;
   if (seq[0][0] % SAMPFREQ == 0) 
      numsamps--;
   othstat=dvec(6);
   f=dmat(NUMLETTERS,NUMLETTERS);
   fastf=dmat(NUMLETTERS,NUMLETTERS);
   fstat=dmat(NUMLETTERS,NUMLETTERS);
   sstat=dmat(NUMLETTERS,NUMLETTERS);
   holdtg=dmat(3,max2(seq[0][0],seq[1][0])+1);
   for(ktrack=0;ktrack<7;ktrack++)
   {
      hold[ktrack]=dmat(numsamps,seq[1][0]);
   }
   holdee=imat(numsamps,seq[1][0]);
   nestim= ivec(NPARAMS);

   gtol= LTOL/10.0;
   oldhood= -1.0*BIGNUM;
   newhood= -0.5*BIGNUM;
   numits=0;

   printf("The output below is the number of EM cycles thus far\n");
   printf("followed by the current parameter estimates:\n");
   while(newhood-oldhood > LTOL) /* continue EM cycles until false */
   {
      for(ktrack=0;ktrack<=NPARAMS;ktrack++)
      {
         if (estim[ktrack]==1 || pars[ktrack] > mask[ktrack][0] || 
            pars[ktrack] < mask[ktrack][1])
         {
            nestim[ktrack]=1;
         }
         else
         {
            nestim[ktrack]=0;
         }
      }
      oldhood=newhood;
      printf("%3d genrate=%8.4f mu=%6.4f r=%6.4f pfast=%6.4f kfast=%6.3f",
      numits,pars[1],pars[2],pars[3],pars[4],pars[5]);
      printf(" tgap=%6.4f\n", pars[6]);
#ifndef PROT
      printf("wigrate= %7.5f\n", pars[7]);
#endif
      newhood= -dubsum(holdtg,hold, holdee,f,fastf,seq,pars);
      printf("newhd= %11.9E oldhd= %11.9E", newhood,oldhood);
      printf(" (lambda/mu =%6.4f)\n", pars[0]);
      /* E-step */
      expect(holdtg,hold, holdee,pars,f, fastf,seq,fstat,sstat,othstat);
      /* following is the M step - what to do for M step depends on 
         parameters to be estimated */
#ifdef PROT
      if (nestim[1]==0)
      {
         if (nestim[5]==0)
         {
            pars[1]=gsearch(0.5*pars[1],pars[1]+0.005, 0.01+
                    1.5*pars[1],fpam,0,0.0,fstat,sstat,othstat,gtol);
         }
         else /* nestim[5]==1, ratio between replacement rates in regions is
                 fixed but slow and fast replacement AMOUNTS are not independent                 so density fspam is needed */
         {
            pars[1]= gsearch(0.5*pars[1], 0.005+pars[1],0.01+ 1.5*pars[1],
                     fspam, 0, pars[5], fstat,sstat,othstat,gtol);
         }
      }   
      if (nestim[5]==0)
      {
         /* Estimate kfast times genrate, get kfast by dividing by genrate */
         pars[5]= gsearch(pars[1],0.005+10.0*pars[1],0.01+50.0*pars[1],fpam,
              0,0.0,sstat,fstat,othstat,gtol)/pars[1];
      }
#else
      estdnasub(nestim,fstat,sstat,pars,gtol);
#endif
      if (nestim[3]==0)
      {
         pars[3]= othstat[5]/(othstat[1]+othstat[2] + othstat[3]*2.0 +
             othstat[4] + othstat[5]);
      }
      if (nestim[2] == 0 && mask[2][0] > pars[2] && mask[2][1] < pars[2] )
      {
         if (nestim[0]==0)
         {
            mulmusearch(pars,seq[0][0]+seq[1][0],fstat,sstat,othstat,gtol);
         }
         else
         {   
            pars[2]= gsearch(0.5*pars[2], 0.005+pars[2], 0.01+1.5*pars[2],
              fmulmu,-seq[0][0]-seq[1][0], pars[0], fstat, sstat,othstat, gtol);         }
      }   
      else
      {   
         if (nestim[0]==0)
         {
            pars[0]= gsearch(0.5*pars[0], pars[0], (1.0+pars[0])/2.0, fmulmu,
              seq[0][0] +seq[1][0], pars[2], fstat, sstat,othstat, gtol);
         }
      }   
      if ( nestim[4]==0)
      {
         pars[4]= othstat[2]/(othstat[1]+othstat[2]);
      }
      if (nestim[6]==0)
      {
         pars[6]= othstat[6]/(2.0+othstat[6]);
      }

      numits++;
   }

   newhood= -dubsum(holdtg,hold, holdee,f,fastf,seq,pars);
   printf("newhd= %11.9E oldhd= %11.9E\n", newhood,oldhood);
   /* now send arc probabilities to output file, later to be turned into
      postscript */
   expectps(holdtg,hold, holdee,pars,f, fastf,seq,fstat,sstat,othstat);

   printf("\n%3d EM iterations were required before log-likelihood was\n",
           numits);
   printf("not improved by at least %8.4E between successive EM ", LTOL);
   printf("iterations\n" );
   printf("gtol=%8.4E (criterion for terminating ", gtol);
   printf("M-step of given EM iteration)\n\n");
   printf("FINAL log-likelihood= %12.10E (", newhood);
   printf("sum over all alignments)\n");
 
   freeivec(nestim,NPARAMS);
   freedvec(othstat,6);
   freedmat(fstat,NUMLETTERS,NUMLETTERS);
   freedmat(sstat,NUMLETTERS,NUMLETTERS);
   freedmat(fastf,NUMLETTERS,NUMLETTERS);
   freedmat(f,NUMLETTERS,NUMLETTERS);
   for(ktrack=0;ktrack<7;ktrack++)
   {
      freedmat(hold[ktrack],numsamps,seq[1][0]);
   }
   freedmat(holdtg,3,max2(seq[0][0],seq[1][0])+1);
   freeimat(holdee,numsamps,seq[1][0]);
}

