 
/* (c) Copyright 1993 by Jeff Thorne.  This program can be copied as long as
   this message is not deleted and the program is distributed free of charge */
 
#include <math.h>
#include <stdio.h>
#include "common.h"
#ifdef TOPS
#include "emh4ps.h"
#endif
#ifdef PROT
#include "prottg.h"
#else
#include "dna.h"
#endif

extern double bfrq[NUMLETTERS+1];

/* The routine expectps() finds arc expectations and outputs arc probabilities
   to a file that can alter be converted to postscript format.  
   It is compiled when preprocessor condition TOPS is
   specified.  Otherwise the routine expect() is compiled. 
   The routine expect() is identical to above but does not output arc probs */
#ifdef TOPS
void 
expectps(holdtg,hold,holdee,pars,f,fastf,seq,fstat,sstat,othstat)
#else
void 
expect(holdtg,hold,holdee,pars,f,fastf,seq,fstat,sstat,othstat)
#endif

/* pars holds the parameter values, "holdtg" and "hold[7]" store some of the arc
   probabilities in the directed graph - "tg" indicates terminal gap  */ 
double *pars,**holdtg,**hold[7];

int **holdee; /* hold some numbers to prevent underflow errors */
double **fstat; /* expected number of fast substitutions of each type */
double **sstat; /* expected number of slow substitutions of each type */
double *othstat; /* expected number for other kinds of stats:
othstat[1]= number of surviving slow links
othstat[2]= number of surviving fast links
othstat[3]= number of extinct links with at least one descendant
othstat[4]= number of extinct links + number newly inserted links
             minus 2*othstats[3]
othstat[5]= number of "independently derived" amino acids in two sequences
             minus number of "independently derived" fragments where 
             there is 1 ind. derived entity per paired entity (i.e. amino acid
             or fragment) and 1 per unpaired alignment entity.  
othstat[6]= number of terminal gap positions */
double **f,**fastf;
int **seq;
{

#ifdef TOPS
   FILE *outfile;
#endif

   int ktrack1,ktrack2,ktrack3,track; 
   int high1,o1hold,n1hold,numsamps;
   int s1hold,s2hold,seq1len,seq2len;
   double **phold ; /* definition here */
   double tgap,r,siterm,i1term,i2term,i3term,i4term,i5term,i6term,i7term;
   double b1hold,b2hold;
   double d1term,d23term,d4term,d5term,d6term,d7term;
   double f1rat,f23rat,f4rat,f7rat;
   double frac1,frac2,frac3,frac4,frac7;
   double mfrac1,mfrac2,mfrac3,mfrac4,mfrac7;
   double **filltg,**fill[7];
   double tempsum,magnify,microfy,magdiag,maghor,magvert;
   double *prvec,genp1,beta,mu,lambda,lmurat;
   double *tf1,*tf2,*tf3,*tf4,*tf5,*tf6,*tf7;
   double *of1,*of2,*of4,*of5,*of7;
   double rof1,rof4,rof7;
   double uf3,uf6;
   double itg2term,itg3term,uptg3,optg2,oftg2,uftg3;
   double tptg3,tptg2,tptg1,tptg0,*tftg0,*tftg1,*tftg2,*tftg3;
   double *tp1,*tp2,*tp3,*tp4,*tp5,*tp6,*tp7;
   double op1,op2,op4,op5,op7;
   double rop1,rop4,rop7;
   double up3,up6;
   int twojumps;
   int *tpee,*upee,*opee,ropee,**pee;
   extern void recreate();
   extern double **dmat();
   extern int **imat();
   extern double *dvec();
   extern void freedmat();
   extern void freeimat();
   extern void freedvec();

#ifdef TOPS
   extern void topsfile();
#endif

   seq1len=seq[0][0];
   seq2len=seq[1][0];

#ifdef TOPS
   outfile=fopen("arcinfo.dat","w");
   fprintf(outfile,"%4d %4d\n",seq1len,seq2len);
   for(ktrack1=1;ktrack1<=seq1len;ktrack1++)
   {
      fprintf(outfile,"%1c",tobase(seq[0][ktrack1]));
      if (ktrack1 %70 ==0) 
         fprintf(outfile,"\n");
   }
   fprintf(outfile,"\n\n");
   for(ktrack1=1;ktrack1<=seq2len;ktrack1++)
   {
      fprintf(outfile,"%1c",tobase(seq[1][ktrack1]));
      if (ktrack1 %70 ==0) 
         fprintf(outfile,"\n");
   }
   fprintf(outfile,"\n\n");
   fprintf(outfile,"%4.2f  is the THRESHOLD BELOW WHICH ARC PROBS NOT SHOWN",
   THRESHOLD);
   fprintf(outfile,"\n>\n");
#endif
   /* set up some useful quantities */
   prvec=dvec(6);
   prvec[0]=pars[3];
   lmurat=pars[0];
   mu=pars[2];
   lambda=lmurat*pars[2];
   beta=(1.0-exp(lambda-mu))/(mu-lambda*exp(lambda-mu)); 
   genp1=exp(-mu)*(1.0-lambda*beta)*(1.0-prvec[0])*lmurat;
   tgap=pars[6];
   prvec[1]= (1.0-pars[4])*genp1;
   prvec[3]= pars[4]*genp1;
   prvec[2]=lambda*beta*(1.0-prvec[0]);
   prvec[4]=(1.0-prvec[0])*(1.0-prvec[0])*(1.0-exp(-mu)-mu*beta)*
   (1.0-lambda*beta)*lmurat;
   prvec[5]=(1.0-lambda*beta)*(1.0-lmurat);
   prvec[6]=pars[6];
   r=prvec[0];

   twojumps= JUMP+JUMP;
   magnify= exp(1.0*twojumps);
   microfy= exp(-1.0*twojumps);
   phold=dmat(6,seq2len);
   for(ktrack1=0;ktrack1<7;ktrack1++)
   {
      fill[ktrack1]=dmat(SAMPFREQ,seq2len);
   }
   pee=imat(SAMPFREQ,seq2len);
   d1term=prvec[1];
   d23term=prvec[2];
   d4term=prvec[4];
   d5term=r;
   d6term=r;
   d7term=prvec[3];
   if (d1term==0.0) 
      f1rat=0.0;
   else 
      f1rat= r/d1term;
   if (d23term==0.0) 
      f23rat=0.0;
   else 
      f23rat=r/d23term;
   if (d4term==0.0) 
      f4rat=0.0;
   else 
      f4rat=r*r/d4term;
   if (d7term==0.0) 
      f7rat=0.0;
   else 
      f7rat=r/d7term;
   numsamps = seq1len/SAMPFREQ;
   if (seq1len% SAMPFREQ == 0) 
      numsamps--;
   filltg=dmat(1,SAMPFREQ);
   for(ktrack1=0;ktrack1<=6;ktrack1++) othstat[ktrack1]=0.0;
   for(ktrack1=0;ktrack1<=NUMLETTERS;ktrack1++)
   {
      for(ktrack2=0;ktrack2<= NUMLETTERS;ktrack2++)
      {
         fstat[ktrack1][ktrack2]=0.0;
         sstat[ktrack1][ktrack2]=0.0; 
      }
   }
   for(ktrack3=numsamps;ktrack3 >=0;ktrack3--)
   {
      /* restore portion of the forward likelihood matrix */
      recreate(holdtg,filltg,prvec,f,fastf,seq,hold,
      holdee,ktrack3,fill,pee); 
      if (ktrack3<numsamps)
      {
         tf1= &fill[0][SAMPFREQ][0];
         tf2= &fill[1][SAMPFREQ][0];
         tf3= &fill[2][SAMPFREQ][0];
         tf4= &fill[3][SAMPFREQ][0];
         tf5= &fill[4][SAMPFREQ][0];
         tf6= &fill[5][SAMPFREQ][0];
         tf7= &fill[6][SAMPFREQ][0];
         tftg1= &filltg[0][SAMPFREQ];
         tftg2= &filltg[1][SAMPFREQ];
         *tftg1= holdtg[1][ktrack3+1];
         *tftg2= holdtg[2][ktrack3+1];
         tpee= &pee[SAMPFREQ][0];
         for(ktrack2=0;ktrack2<=seq2len;ktrack2++)
         {
            *tf1=hold[0][ktrack3+1][ktrack2];
            *tf2=hold[1][ktrack3+1][ktrack2];
            *tf3=hold[2][ktrack3+1][ktrack2];
            *tf4=hold[3][ktrack3+1][ktrack2];
            *tf5=hold[4][ktrack3+1][ktrack2];
            *tf6=hold[5][ktrack3+1][ktrack2];
            *tf7=hold[6][ktrack3+1][ktrack2];
            *tpee=holdee[ktrack3+1][ktrack2];
            tf1++;
            tf2++;
            tf3++;
            tf4++;
            tf5++;
            tf6++;
            tf7++;
            tpee++;
         }
      } 
      else /* ktrack3=numsamps */ 
      {
         track=seq1len-numsamps*SAMPFREQ;
         ktrack2=seq2len;
         ktrack1=seq1len;

         tf1= &fill[0][track][ktrack2];
         tf2= &fill[1][track][ktrack2];
         tf3= &fill[2][track][ktrack2];
         tf4= &fill[3][track][ktrack2];
         tf5= &fill[4][track][ktrack2];
         tf6= &fill[5][track][ktrack2];
         tf7= &fill[6][track][ktrack2];
         tftg1= &filltg[0][track];
         tftg2= &filltg[1][track];
         tftg3= &holdtg[3][ktrack2];
         tp1= &phold[0][ktrack2];
         tp2= &phold[1][ktrack2];
         tp3= &phold[2][ktrack2];
         tp4= &phold[3][ktrack2];
         tp5= &phold[4][ktrack2];
         tp6= &phold[5][ktrack2];
         tp7= &phold[6][ktrack2];
         tempsum = *tf1+ *tf2 + *tf3 + *tf4 + *tf5+ *tf6+ *tf7+ *tftg2+ *tftg3;
         *tp1= *tf1/tempsum;
         *tp2= *tf2/tempsum;
         *tp3= *tf3/tempsum;
         *tp4= *tf4/tempsum;
         *tp5= *tf5/tempsum;
         *tp6= *tf6/tempsum;
         *tp7= *tf7/tempsum;
         tptg2= *tftg2/tempsum;
         tptg3= *tftg3/tempsum;
#ifdef TOPS
         topsfile(outfile,ktrack1,ktrack2,3,*tp7); 
         topsfile(outfile,ktrack1,ktrack2,2,*tp1); 
         topsfile(outfile,ktrack1,ktrack2,5,*tp2); 
         topsfile(outfile,ktrack1,ktrack2,7,tptg2); 
         topsfile(outfile,ktrack1,ktrack2,0,*tp3); 
         topsfile(outfile,ktrack1,ktrack2,8,tptg3); 
         topsfile(outfile,ktrack1,ktrack2,4,*tp4); 
         topsfile(outfile,ktrack1,ktrack2,6,*tp5); 
         topsfile(outfile,ktrack1,ktrack2,1,*tp6); 
#endif
         tpee= &pee[track][ktrack2];
         
         s1hold=seq[0][ktrack1];
         s2hold=seq[1][ktrack2];
         fstat[s1hold][s2hold] += *tp7;
         sstat[s1hold][s2hold] += *tp1; 
         othstat[1] += (*tp1);
         othstat[2] += (*tp7);
         othstat[3] += (*tp4);
         othstat[4] += (*tp2)+(*tp3);
         othstat[5] += (*tp5)+(*tp6);
         othstat[6] += tptg2+tptg3;
         for(ktrack2=seq2len-1; ktrack2 >=0;ktrack2--)
         {
            uftg3= *tftg3;
            uf3= *tf3;
            uf6= *tf6;
            tf1--;
            tf2--;
            tf3--;
            tf4--;
            tf5--;
            tf6--;
            tf7--;
            tftg3--;
            uptg3= tptg3;
            up3= *tp3;
            up6= *tp6;
            tp1--;
            tp2--;
            tp3--;
            tp4--;
            tp5--;
            tp6--;
            tp7--;
            upee=tpee;
            tpee--;
            magvert=bfrq[s2hold];
            if (*upee > *tpee) 
               magvert *=magnify;
            else 
            {
               if (*upee < *tpee) 
                  magvert *=microfy;
            }
            if (up3 == 0.0) 
            {
               i3term=0.0;
            } 
            else
            {
               i3term=magvert*up3*d23term/ uf3; 
            }
            if (up6 ==0.0)
            {
               i6term=0.0;
            }
            else
            {
               i6term=magvert*up6*d6term/ uf6;
            }
            if (uptg3==0.0) 
            {
               itg3term=0.0;
            }
            else
            {
               itg3term = magvert*uptg3*0.5*tgap/uftg3;
            }
            *tp1= *tf1*(i3term+itg3term);
            othstat[1] += *tp1;
            *tp2= *tf2*itg3term;
            mfrac3= *tf3*(i3term+itg3term);
            frac3=  *tf3*i3term*f23rat;
            *tp3= frac3+mfrac3;
            othstat[4] += *tp2+mfrac3;
            *tp4= *tf4*(i3term+itg3term+i6term);
            othstat[3] += *tp4;
            *tp5= *tf5*(i3term+itg3term);
            *tp6= *tf6* (i3term+itg3term+i6term);
            *tp7= *tf7*(itg3term+i3term);
            tptg3= *tftg3*2.0*itg3term;
            othstat[6] += tptg3;
            othstat[2] += *tp7;
            othstat[5] +=frac3+(*tp5) + (*tp6);
#ifdef TOPS
            topsfile(outfile,ktrack1,ktrack2,3,*tp7);
            topsfile(outfile,ktrack1,ktrack2,2,*tp1);
            topsfile(outfile,ktrack1,ktrack2,0,*tp3);
            topsfile(outfile,ktrack1,ktrack2,8,tptg3);
            topsfile(outfile,ktrack1,ktrack2,4,*tp4);
            topsfile(outfile,ktrack1,ktrack2,6,*tp5);
            topsfile(outfile,ktrack1,ktrack2,1,*tp6);
            topsfile(outfile,ktrack1,ktrack2,5,*tp2);
#endif
            if (ktrack2>0)
            {
               s2hold= seq[1][ktrack2];
               fstat[s1hold][s2hold] += *tp7;
               sstat[s1hold][s2hold] += *tp1;
            }
            else
            {
               tptg1= *tftg1*(i3term+itg3term);
               othstat[6] += tptg1;
#ifdef TOPS
               topsfile(outfile,ktrack1,ktrack2,7,tptg1);
#endif
            } 
         } 
      }
      high1=min2(seq[0][0]-1,(ktrack3+1)*SAMPFREQ-1);
      n1hold=seq[0][high1+1];

      for(ktrack1=high1; ktrack1 >=ktrack3* SAMPFREQ; ktrack1--)
      {
         s2hold=seq[1][seq2len];
         b2hold=bfrq[s2hold];
         o1hold=n1hold;
         s1hold=o1hold;
         b1hold=bfrq[s1hold];
         if (ktrack1 != 0) 
            n1hold=seq[0][ktrack1];
         else
         {
            tftg0= &holdtg[0][seq2len];
         }
         track=ktrack1-ktrack3*SAMPFREQ;
         ktrack2=seq[1][0];
         tpee= &pee[track][seq2len];
         opee= &pee[track+1][seq2len];

         
         tf1= &fill[0][track][seq2len];
         tf2= &fill[1][track][seq2len];
         tf3= &fill[2][track][seq2len];
         tf4= &fill[3][track][seq2len];
         tf5= &fill[4][track][seq2len];
         tf6= &fill[5][track][seq2len];
         tf7= &fill[6][track][seq2len];
         tftg1= &filltg[0][track];
         tftg2= &filltg[1][track];
         oftg2 = *(tftg2+1); 
         of1= &fill[0][track+1][seq2len];
         of2= &fill[1][track+1][seq2len];
         of4= &fill[3][track+1][seq2len];
         of5= &fill[4][track+1][seq2len];
         of7= &fill[6][track+1][seq2len];

         tp1= &phold[0][seq2len];
         tp2= &phold[1][seq2len];
         tp3= &phold[2][seq2len];
         tp4= &phold[3][seq2len];
         tp5= &phold[4][seq2len];
         tp6= &phold[5][seq2len];
         tp7= &phold[6][seq2len];
         optg2= tptg2;
         op1= *tp1;
         op2= *tp2;
         op4= *tp4;
         op5= *tp5;
         op7= *tp7;

         maghor=b1hold;
         if (*opee> *tpee) 
            maghor *=magnify;
         else 
         {
            if (*opee < *tpee) 
               maghor *=microfy;
         }
         if (op2==0.0)
         {
            i2term=0.0;
         }
         else
         {
            i2term=maghor*op2*d23term/ (*of2) ; 
         }
         if (op5==0.0)
         {
            i5term=0.0;
         }
         else
         {
            i5term=maghor*op5*d5term/ (*of5);
         }
         if (optg2==0.0)
         {
            itg2term=0.0;
         }
         else
         {
            itg2term=maghor*optg2*tgap*0.5/oftg2;
         } 
         *tp1= (*tf1)*(itg2term+i2term);
         othstat[1] += *tp1;
         frac2=  (*tf2)*i2term*f23rat;
         mfrac2= (*tf2)*(itg2term+i2term);
         *tp2=frac2+mfrac2;
         *tp3= (*tf3)*(itg2term+i2term);
         othstat[4] += mfrac2 +*tp3;
         *tp4= (*tf4)*(itg2term+i2term+i5term);
         othstat[3] += *tp4;
         *tp5= (*tf5)* (i2term+i5term+itg2term);
         *tp6= (*tf6)* (itg2term+i2term);
         *tp7= (*tf7)*(itg2term+i2term);
         othstat[2] += *tp7;
         othstat[5] +=frac2+(*tp5)+(*tp6);
         tptg2= 2.0*itg2term*(*tftg2);
         othstat[6] += tptg2;
#ifdef TOPS
         topsfile(outfile,ktrack1,ktrack2,3,*tp7);
         topsfile(outfile,ktrack1,ktrack2,2,*tp1);
         topsfile(outfile,ktrack1,ktrack2,5,*tp2);
         topsfile(outfile,ktrack1,ktrack2,7,tptg2);
         topsfile(outfile,ktrack1,ktrack2,4,*tp4);
         topsfile(outfile,ktrack1,ktrack2,6,*tp5);
         topsfile(outfile,ktrack1,ktrack2,1,*tp6);
         topsfile(outfile,ktrack1,ktrack2,0,*tp3);
#endif
         if (ktrack1>0)
         {
            s1hold=n1hold;
            fstat[s1hold][s2hold] += *tp7;
            sstat[s1hold][s2hold] += *tp1;
            s1hold=o1hold;
         }
         else
         {
            tptg0= (*tftg0)*(i2term+itg2term);
            othstat[6] += tptg0;
#ifdef TOPS
            topsfile(outfile,ktrack1,ktrack2,8,tptg0);
#endif
         }
         for(ktrack2=seq2len-1;ktrack2 >=0;ktrack2--)
         {
            uf3= *tf3;
            uf6= *tf6;
            rof1= *of1;
            rof4= *of4;
            rof7= *of7;
            of1--;
            of2--;
            of4--;
            of5--;
            of7--;
            tf1--;
            tf2--;
            tf3--;
            tf4--;
            tf5--;
            tf6--;
            tf7--;
            rop1= op1;
            rop4= op4;
            rop7= op7;
            up3= *tp3;
            up6= *tp6;
            tp1--;
            tp2--;
            tp3--;
            tp4--;
            tp5--;
            tp6--;
            tp7--;
            op1= *tp1;
            op2= *tp2;
            op4= *tp4;
            op5= *tp5;
            op7= *tp7;

            upee=tpee;
            tpee--;
            ropee= *opee;
            opee--;

            if (rop1 ==0.0) 
               i1term=0.0;
            else 
               i1term=rop1*d1term*f[s1hold][s2hold]/ rof1; 

            if (op2 ==0.0) 
               i2term=0.0;
            else 
               i2term=op2*d23term*b1hold/ (*of2);

            if (up3==0.0) 
               i3term=0.0;
            else 
               i3term=up3*d23term* b2hold/ uf3; 

            if (rop4==0.0) 
               i4term=0.0;
            else 
               i4term=rop4*d4term*b1hold*b2hold/ rof4;

            if (op5==0.0) 
               i5term=0.0;
            else 
               i5term=op5*d5term*b1hold/ (*of5);

            if (up6 ==0.0) 
               i6term=0.0;
            else 
               i6term= up6*d6term* b2hold/ uf6;

            if (rop7 ==0.0) 
               i7term=0.0;
            else 
               i7term=rop7*d7term*fastf[s1hold][s2hold]/rof7; 

            if (*tpee != ropee)
            {
               if (ropee > *tpee) 
                  magdiag=magnify;
               else 
                  magdiag=microfy;
               i1term *= magdiag;
               i4term *= magdiag;
               i7term *= magdiag;
            }
            if (*opee != *tpee)
            {
               if (*opee > *tpee) 
                  maghor=magnify;
               else 
                  maghor=microfy;
               i2term *= maghor;
               i5term *= maghor;
            }
            if (*upee != *tpee)
            {
               if (*upee > *tpee) 
                  magvert=magnify;
               else 
                  magvert= microfy;
               i3term *= magvert;
               i6term *= magvert;
            } 
            /* travelling 3' to 5', calculate the probability of the  
               ktrack1,ktrack2 position being in the alignment and
               in a specific category */

            if (ktrack1==0 && ktrack2==0) 
            {
               *tp1=0.0;
               *tp2=0.0;
               *tp3=0.0;
               *tp4=0.0;
               *tp5=0.0;
               *tp6=0.0;
               *tp7=0.0;
            }
            else
            {
               siterm=i1term+i2term+i3term+i4term+i7term;
               if (ktrack2==0)
               {
                  tptg1 += *tftg1*siterm;
                  othstat[6] += tptg1;
#ifdef TOPS
                  topsfile(outfile,ktrack1,ktrack2,7,tptg1);
#endif
               }
               if (ktrack1==0)
               {
                  tftg0--;
                  tptg0 += *tftg0*siterm;
                  othstat[6]  += tptg0;
#ifdef TOPS
                  topsfile(outfile,ktrack1,ktrack2,8,tptg0);
#endif
               }
               frac1=  (*tf1)*i1term*f1rat;
               mfrac1= (*tf1)* siterm;
               frac2= (*tf2)*i2term*f23rat;
               mfrac2= (*tf2)*(siterm-i3term);
               frac3= (*tf3)*i3term*f23rat;
               mfrac3=(*tf3)* siterm;
               frac4= (*tf4)*i4term*f4rat;
               mfrac4=(*tf4)*(siterm+i5term+i6term);
               frac7= (*tf7)*i7term*f7rat;
               mfrac7=(*tf7)*siterm;

               *tp1 = frac1+mfrac1;
               *tp2=frac2+mfrac2;
               *tp3 =frac3+mfrac3;
               *tp4=frac4+mfrac4;
               *tp5= (*tf5)*(siterm+i5term);
               *tp6= (*tf6)* (siterm+i6term);
               *tp7=frac7+mfrac7;
               othstat[1] += mfrac1;
               othstat[2] += mfrac7;
               othstat[3] += mfrac4;
               othstat[4] += mfrac3+mfrac2;
               othstat[5] +=frac1+frac2+frac3+2.0*frac4+frac7+(*tp5)+(*tp6);
#ifdef TOPS
               topsfile(outfile,ktrack1,ktrack2,5,*tp2);
               topsfile(outfile,ktrack1,ktrack2,0,*tp3);
               topsfile(outfile,ktrack1,ktrack2,3,*tp7);
               topsfile(outfile,ktrack1,ktrack2,2,*tp1);
               topsfile(outfile,ktrack1,ktrack2,6,*tp5);
               topsfile(outfile,ktrack1,ktrack2,4,*tp4);
               topsfile(outfile,ktrack1,ktrack2,1,*tp6);
#endif
            }
            if (ktrack2 >0)
            {
               s2hold= seq[1][ktrack2];
               b2hold=bfrq[s2hold];
               if (ktrack1>0)
               {
                  s1hold= n1hold;
                  fstat[s1hold][s2hold] += *tp7;
                  sstat[s1hold][s2hold] += *tp1;
                  s1hold=o1hold;
               }
            }
         }
      }
   }
#ifdef TOPS
   fprintf(outfile,"0 0 0 -1.0\n");
   fclose(outfile);
#endif
   freedmat(filltg,1,SAMPFREQ);
   freedvec(prvec,6);
   freeimat(pee,SAMPFREQ,seq2len);
   for(ktrack1=0;ktrack1<7;ktrack1++)
   {
      freedmat(fill[ktrack1],SAMPFREQ,seq2len);
   }
   freedmat(phold,6,seq2len);
}
