 
/* (c) Copyright 1993 by Jeff Thorne.  This program can be copied as long as
   this message is not deleted and the program is distributed free of charge */
 
#include <stdio.h>
#include "common.h"
#ifndef PROT
#include "dna.h"
#endif

/* brackets and then does golden section search  */
double 
gsearch(lpt,mpt,rpt,inifunc,parflag,othpar,othsubstat,substat,othstat,gtol)
double lpt,mpt,rpt,gtol;
double (*inifunc)();
double othpar,**othsubstat, **substat,*othstat;
int parflag;
{
   int numits;
   double lhd,mhd,rhd,newpt,newhd,lftint,rtint;

   lpt= max2(lpt,SMALLNUM);
   mpt= max2(mpt,SMALLNUM);
   rpt= max2(rpt,SMALLNUM);

   if (parflag > 0)
   {
      lpt=min2(lpt,1.0-SMALLNUM);
      mpt=min2(mpt,1.0-SMALLNUM);
      rpt=min2(rpt,1.0-SMALLNUM);
   }
   else
   {
      lpt=min2(lpt,BIGNUM);
      mpt=min2(mpt,BIGNUM);
      rpt=min2(rpt,BIGNUM);
   }
 
   lhd= (*inifunc)(lpt,parflag,othpar,othsubstat,substat,othstat);
   mhd= (*inifunc)(mpt,parflag,othpar,othsubstat,substat,othstat);
   rhd= (*inifunc)(rpt,parflag,othpar,othsubstat,substat,othstat);
   

   numits=0;
   if (mhd != max3(lhd,mhd,rhd))
   {
      while ( mhd != max3(lhd,mhd,rhd) && numits <=10)
      {
         if (lhd==max3(lhd,mhd,rhd))
         {
            rpt=mpt;
            rhd=mhd;
            mpt=lpt;
            mhd=lhd;
            lpt= max2(SMALLNUM, lpt/10.0);
            lhd = (*inifunc)(lpt,parflag,othpar,othsubstat,substat,othstat);  
         }
         else 
         {
            lpt=mpt;
            lhd=mhd;
            mpt=rpt;
            mhd=rhd;
            if (parflag > 0) 
               rpt= min2( (1.0+rpt)/2.0, 1.0-SMALLNUM);
            else 
               rpt = min2(BIGNUM, rpt*10.0);
            rhd = (*inifunc)(rpt,parflag,othpar,othsubstat,substat,othstat);  
         }
         numits++;
      }
   }
   if (numits > 10)
   {
      printf("Trouble: mhd not in middle !!\n");
      printf("Can't Recover\n");
      printf("lpt= %8.6E lhd= %8.6E\n",lpt,lhd);
      printf("mpt= %8.6E mhd= %8.6E\n",mpt,mhd);
      printf("rpt= %8.6E rhd= %8.6E\n\n",rpt,rhd);
   } 
   else 
   {
      while (mhd-rhd > gtol || mhd-lhd > gtol)
      {
         lftint= mpt-lpt;
         rtint=rpt-mpt;

         if (lftint>rtint)
         {
            newpt= mpt-0.38197*lftint;
            newhd= (*inifunc)(newpt,parflag,othpar,othsubstat,substat,othstat);
            if (newhd>mhd)
            {
               rpt=mpt;
               rhd=mhd;
               mpt=newpt;
               mhd=newhd;
            }
            else
            {
               lpt=newpt;
               lhd=newhd;
            }
         }  
         else
         {
            newpt=mpt+0.38197*rtint;
            newhd= (*inifunc)(newpt,parflag,othpar,othsubstat,substat,othstat);
            if (newhd>mhd)
            {
               lpt=mpt;
               lhd=mhd;
               mpt=newpt;
               mhd=newhd;
            }
            else
            {
               rpt=newpt;
               rhd=newhd;
            }
         }  
      }
   }
   return(mpt);
}

/* searches for best joint estimate of deletion parameter and ratio of
   insertion rate to deletion rate */
void 
mulmusearch(pars,parflag,othsubstat, substat,othstat,gtol)
double *pars,**othsubstat,**substat,*othstat,gtol;
int parflag;
{
   double gsearch();
   extern double fmulmu();

   double oldlmu,newlmu,oldmu,newmu,oldhood,newhood;

   oldlmu=pars[0];
   oldmu=pars[2];
   oldhood= -1.0*BIGNUM;
   newhood= -0.5*BIGNUM;
   while (newhood-oldhood>gtol)
   {
      oldhood=newhood;
      newlmu=gsearch(oldlmu/2.0,oldlmu,(1.0+oldlmu)/2.0,
             fmulmu,parflag,oldmu,othsubstat,substat, othstat,gtol);
      newmu= gsearch(oldmu/2.0,0.005+oldmu,0.01+1.5*oldmu,fmulmu,-1*parflag,
              newlmu, othsubstat,substat,othstat,gtol);
      newhood=fmulmu(newlmu,parflag,newmu,othsubstat,substat,othstat);
      oldlmu=newlmu;
      oldmu=newmu;
   }
   pars[0]=newlmu;
   pars[2]=newmu;
}

#ifndef PROT

extern double bfrq[];
extern double gfrq[];
extern int group[];

/* searches for best estimates of specified
   2 or 3 of the substitution parameters */
void 
g2subsearch( parflag,othsubstat, substat,pars,gtol)
double **othsubstat,**substat,*pars,gtol;
int parflag;
{
   double gsearch();
   extern double fslsub();
   extern double ktofsub();

   double oldgen, oldwig,  oldk, oldhood,newhood;

   oldk=pars[5];
   oldgen=pars[1];
   oldwig=pars[7];
   oldhood= -1.0*BIGNUM;
   newhood= -0.5*BIGNUM;
   while (newhood-oldhood>gtol)
   {
      oldhood=newhood;

      if (parflag != -5)
      {
         pars[5]=gsearch(0.5*(1.0+oldk),0.00005+oldk,0.01+1.5*oldk,
             ktofsub,-5,0.0,othsubstat,substat, pars,gtol);
         pars[5]= max2(pars[5],1.0);
      }
      else pars[5]=oldk;
      if (parflag != -1)
      {
         pars[1]=gsearch(0.5*oldgen,0.005+oldgen,0.01+1.5*oldgen,
             fslsub,-1,0.0,othsubstat,substat, pars,gtol);
      }
      else pars[1]=oldgen;
      if (parflag != -7)
      {
         pars[7]=gsearch(0.5*oldwig,0.005+oldwig,0.01+1.5*oldwig,
             fslsub,-7,0.0,othsubstat,substat, pars,gtol);
      }
      else pars[7]=oldwig;

      newhood= fslsub(pars[1],-1, 0.0, othsubstat, substat, pars);
      oldgen=pars[1];
      oldwig=pars[7];
      oldk=pars[5];
   }
}

/* Determines what substitution parameters need to be estimated and then
   acts accordingly */
void  
estdnasub(estim,fstat,sstat,pars,gtol)
double **fstat, **sstat, *pars, gtol;
int *estim;
{
   extern double fsub();
   extern double fslsub();
   extern double ktofsub();

   void g2subsearch();


   if (estim[5]==0)
   {
      if (estim[7]==0)
      {
         if (estim[1]==0)
         {
            g2subsearch( -50, fstat, sstat,pars,gtol);
         }
         else /* estim[1]==1 */
         {
            g2subsearch( -1, fstat, sstat,pars,gtol);
         
         }
      }
      else /* estim[7]==1 */
      {
         if (estim[1]==0)
         {
            g2subsearch( -7, fstat, sstat,pars,gtol);
         }
         else /* estim[1]==1 */
         {
            pars[5]=gsearch(0.5*(1.0+pars[5]),0.00005+pars[5],0.01+1.5*pars[5],
            ktofsub,-5,0.0,fstat,sstat, pars,gtol);
            pars[5]= max2(pars[5], 1.0);
         }
      }
   }
   else /* estim[5]==1 */
   {
      if (estim[7]==0)
      {
         if (estim[1]==0)
         {
            g2subsearch( -5, fstat, sstat,pars,gtol);
         }
         else /* estim[1]==1 */
         {
            pars[7]= gsearch(0.5*pars[7],0.005+pars[7],0.01+1.5*pars[7], 
                     fslsub, -7, 0.0, fstat,sstat,pars,gtol); 
         }   
      }
      else /* estim[7]==1 */
      {
         if (estim[1]==0)
         {
            pars[1]= gsearch(0.5*pars[1], 0.005+pars[1], 0.01+1.5*pars[1], 
                     fslsub, -1, 0.0, fstat, sstat, pars,gtol);
         }
         /* else  estim[1]==1 */
      }
   }
}
 

#endif
