
/* (c) Copyright 1993 by Jeff Thorne.  This program can be copied as long as
   this message is not deleted and the program is distributed free of charge */

/* many of the routines that are necessary for comparing pairs of DNA
   or protein sequences.  Routines for estimating evolutionary parameters,
   inferring alignments, etc. */ 

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#ifdef PROT
#include "prottg.h" 

/* Eigenvalues and Eigenvectors for Substitution Matrix */
double lneval[NUMLETTERS];
double evec[NUMLETTERS][NUMLETTERS],ievec[NUMLETTERS][NUMLETTERS];

#else
#include "dna.h"
double gfrq[NUMGROUPS];
int group[NUMLETTERS];
#endif

/* Frequencies of amino acids or bases */
double bfrq[NUMLETTERS+1];
/* parameters for alignment, sloppy programming, shouldn't be global */
int alilength,*ali;

main()
{
   int fbound; /* flag determining whether fragment boundaries are
                  inferred in alignment */
   int **seq;  /* The matrix which holds the actual sequence info. */
   int *estim; /* contains info. about the parameters to be estimated */
   char *name1,*name2;
   double **mask; /* contains highest and lowest reasonable param values */
   double alihood; /* alignment likelihood */
   double pars[NPARAMS+1]; /* evolutionary parameters */ 
#ifdef CONS
   extern void docons(); /* routine for finding evolutionary distance 
                            under constraints */
#else
   extern void domin(); /* Calls EM routine for parameter estimation */
#endif
   extern double **dmat();
   extern int **imat();
   extern int *ivec();
   extern void freedmat();
   extern void freeimat();
   extern void freeivec();
   extern void readseq(); /* reads in the sequences */
   void dofrq(); /* sets up the PAM matrix or DNA stuff */
   void doderivs(); /* Approx. standard deviations of parameter estimates */
   extern void printali(); /* print alignment */
   double doali(); /* controlling routine for alignment inference */
   /* extern void unrelated(); */ /* Probabilities of alignments */
   extern domask();  /*  reads a file containing users objectives  - i.e., what
                         parameters should be estimated ? */
   
   /* Meaning of the pars array:
   pars[0]=  ratio of lambda (insertion) to mu  (deletion)
   pars[1]= genrate  ("general" substitution rate)
   pars[2]= mu (deletion amount) 
   pars[3]= r (fragment size parameter) 
   pars[4]= pfast  (prob. of fast fragment)
   pars[5]= kfast (ratio of rates in fast to slow fragments)
   pars[6]= tgap (terminal gap parameter)
   pars[7] = wigrate ("within-group" substitution rate - DNA sequences only) */

   estim=ivec(NPARAMS);
   mask=dmat(NPARAMS,1);
   name1= (char *) malloc((unsigned) sizeof(char)*81);
   name2= (char *) malloc((unsigned) sizeof(char)*81);

   domask(pars, estim, mask);
   printf("\nBANDWIDTH= %6d SAMPFREQ= %6d\n", BANDWIDTH,SAMPFREQ);
   seq=imat(1,MAXSEQLENGTH);
   readseq(seq,name1,name2);
   ali=ivec(seq[0][0]+seq[1][0]);
   dofrq(seq);
#ifdef CONS
   docons(seq,pars,estim,mask,name1,name2);
#else
   domin(seq,pars,estim,mask);
   doderivs(seq,pars,estim,mask);
#endif

   /* unrelated(seq,pars, LTOL/10.0);  */ 
   fbound=1;

   printf("Best alignment obtained by ");
   if ( (fbound % 2) == 1) 
   {
      printf("summing over fragment boundaries\n");
   }
   else 
   {
      printf("inferring individual fragment boundaries\n");
   }
   alihood= doali(seq,pars,fbound);
   printf("negative log-likelihood of best alignment= %10.6f\n", alihood);
   printf("A most likely alignment -- there may be ties -- is:\n"); 
   printali(seq);
   free((char*) (name1));
   free((char*) (name2));
   freeivec(ali,seq[0][0]+seq[1][0]);
   freeimat(seq,1,MAXSEQLENGTH);
   freeivec(estim,NPARAMS);
   freedmat(mask,NPARAMS,1);
}

void 
doderivs(seq,pars,estim, mask) 
            /* Approximate Standard Deviations of Estimates */
int **seq, *estim;
double  **mask;
double *pars;
{
   double dubsum(); /* calculate log-likelihood of sequences */
   extern void invert(); /* invert matrix */
   extern double **dmat();
   extern int **imat();
   extern int *ivec();
   extern double *dvec();
   extern void freedmat();
   extern void freeimat();
   extern void freeivec();
   extern void freedvec();

   int ktrack1,ktrack2, nvary;
   double tcorr, mhood,**hihi,**hilo,**lolo,*hino,*lono,*pvec,*inc;
   double **invmat,**derivmat;
   double **f,**fastf;
   /* Below are not important here but are required by dubsum() anyhow */
   double **hold[NPARAMS+1],**holdtg;
   int **holdee;
   int *finmask,numsamps; 
   int dum1,dum2,stopflag,nseqs;

   FILE *thisfar; /* file that tells which sequences are being analyzed */
   FILE *estfile; /* file for estimates of parameters */
   FILE *testseq; /* file with the sequences of interest */
 
   finmask= ivec(NPARAMS);
   numsamps=seq[0][0]/SAMPFREQ;
   if ((seq[0][0]%SAMPFREQ) == 0) 
      numsamps--;
   f=dmat(NUMLETTERS,NUMLETTERS);
   fastf=dmat(NUMLETTERS,NUMLETTERS);
   for (ktrack1=0;ktrack1<=NPARAMS;ktrack1++)
   {
      hold[ktrack1]= dmat(numsamps,seq[1][0]);
   }
   holdtg= dmat(3,max2(seq[0][0],seq[1][0])+1);
   holdee=imat(numsamps,seq[1][0]);

   pvec= dvec(NPARAMS);

   nvary=0;
   for (ktrack1=0;ktrack1<=NPARAMS;ktrack1++) 
   {
      pvec[ktrack1]=pars[ktrack1];
      if (mask[ktrack1][0] < pars[ktrack1] || mask[ktrack1][1] > pars[ktrack1] 
         || estim[ktrack1]==1) 
      {
         finmask[ktrack1]= -1;
      }
      else 
      {
         finmask[ktrack1]= nvary;
         nvary++;
      }
   }
   nvary--;

   if (nvary >=0)
   {
      inc=dvec(nvary);
      hihi=dmat(nvary,nvary);
      lolo=dmat(nvary,nvary);
      hilo=dmat(nvary,nvary);
      hino = dvec(nvary);
      lono = dvec(nvary);
      invmat=dmat(nvary,nvary);
      derivmat=dmat(nvary,nvary);
      for (ktrack1=0;ktrack1<=NPARAMS;ktrack1++)
      {
         if (finmask[ktrack1] >=0)
         {
            if (ktrack1==0 || ktrack1==3 || ktrack1==4 || ktrack1==6)
            {
               inc[finmask[ktrack1]]=0.01*min2(pvec[ktrack1],1.0-pvec[ktrack1]);
            }
            else
            {
               inc[finmask[ktrack1]]= 0.01*pvec[ktrack1];
            }
         }
      }

      mhood= dubsum(holdtg,hold, holdee,f,fastf,seq,pvec);

      for (ktrack1=0;ktrack1<=NPARAMS;ktrack1++) 
      {
         if (finmask[ktrack1] >=0)
         {
            pvec[ktrack1] += inc[finmask[ktrack1]];
            hino[finmask[ktrack1]] = 
               dubsum(holdtg,hold, holdee,f,fastf,seq,pvec);
            for (ktrack2=0;ktrack2<=NPARAMS;ktrack2++) 
            {
               if (finmask[ktrack2] >= 0)
               {
                  pvec[ktrack2]  += inc[finmask[ktrack2]];
                  if (ktrack2 > ktrack1)
                  {
                     hihi[finmask[ktrack1]][finmask[ktrack2]]= 
                        dubsum(holdtg,hold,holdee,f,fastf,seq,pvec);
                  }
                  pvec[ktrack2] -= 2.0 * inc[finmask[ktrack2]];
                  if (ktrack1 != ktrack2) 
                  {
                     hilo[finmask[ktrack1]][finmask[ ktrack2]]= 
                        dubsum(holdtg,hold, holdee,f,fastf,seq,pvec);
                  }
                  pvec[ktrack2] += inc[finmask[ktrack2]];
               }
            }
            pvec[ktrack1] -= 2.0 * inc[finmask[ktrack1]];
            lono[finmask[ktrack1]]=dubsum(holdtg,hold, holdee,f,fastf,seq,pvec);
            for (ktrack2=0;ktrack2 <= NPARAMS;ktrack2++) 
            {
               if (finmask[ktrack2] >= 0)
               {
                  pvec[ktrack2] -= inc[finmask[ktrack2]];
                  if (ktrack2 > ktrack1)
                  {
                     lolo[finmask[ktrack1]][finmask[ktrack2]] =
                        dubsum(holdtg,hold,holdee,f,fastf,seq,pvec);
                  }
                  pvec[ktrack2] += inc[finmask[ktrack2]];
               }
            }
            pvec[ktrack1] += inc[finmask[ktrack1]];
         }
      }

      for (ktrack1=0;ktrack1<=NPARAMS;ktrack1++) 
      {
         if (finmask[ktrack1] >=0)
         {
            derivmat[finmask[ktrack1]][finmask[ktrack1]] = 
               (float) ( (hino[finmask[ktrack1]]  -2.0*mhood + 
               lono[finmask[ktrack1]])/ (inc[finmask[ktrack1]]*
               inc[finmask[ktrack1]]) );

            for (ktrack2=ktrack1+1;ktrack2<=NPARAMS;ktrack2++) 
            {
               if (finmask[ktrack2] >=0)
               {
                  derivmat[finmask[ktrack1]][finmask[ktrack2]] = (float) 
                     ( ( hihi[finmask[ktrack1]][finmask[ktrack2]] -
                     hilo[finmask[ktrack1]][finmask[ktrack2]] - 
                     hilo[finmask[ ktrack2]][ finmask[ktrack1]] + 
                     lolo[finmask[ktrack1]][ finmask[ktrack2]]) / 
                     (4.0*inc[finmask[ktrack1]]*inc[finmask[ ktrack2]]) );
                  derivmat[finmask[ktrack2]][finmask[ktrack1]] = 
                     derivmat[finmask[ktrack1]][finmask[ktrack2]];
               }
            }
         }
      }

      for (ktrack1=0;ktrack1<=nvary;ktrack1++) 
      {
         for (ktrack2=0;ktrack2<=nvary;ktrack2++) 
            invmat[ktrack1][ktrack2] = derivmat[ktrack1][ktrack2];
      }
      invert(invmat,nvary); 

   }

   printf("\nFinal (i.e., maximum likelihood) estimates...\n"); 
   printf("PARAMETER:\n");
   printf("0-(lambda/mu) 1-genrate  2-mu  3-r  4-pfast  5-kfast 6-tgap\n");
#ifndef PROT
   printf("7-wigrate\n");
#endif
   printf("\n");
   for (ktrack1=0;ktrack1<= NPARAMS; ktrack1++) 
   {
      for (ktrack2=ktrack1+1;ktrack2<= NPARAMS; ktrack2++)
      {
         printf("estimated correlation between %1d and %1d is ", 
                 ktrack1, ktrack2);

         if (finmask[ktrack1] >= 0 && finmask[ktrack2]>=0)
         {
            tcorr= invmat[finmask[ktrack1]][finmask[ktrack1]]* 
               invmat[ finmask[ktrack2]][finmask[ktrack2]];
            if (tcorr > 0.0)
            {
               printf("%10.5E\n" , invmat[finmask[ktrack1]][finmask[ktrack2]]/
                      sqrt(tcorr) );
            }
            else
            {
               printf("nonsense (Boundary problem ?).\n");
            }
         }
         else
         {
            printf("not calculated.\n");
         }
      }
   }
   printf("\n");

   
   thisfar= fopen("thisfar", "r");
   fscanf(thisfar,"%d%d%d", &dum1, &dum2, &stopflag);
   fclose(thisfar);

   testseq = fopen("testseq","r");
   fscanf(testseq,"%d", &nseqs);
   fclose(testseq);

   if (dum1==1 && dum2==2)
   {
      estfile= fopen("seqs.est", "w");
      fprintf(estfile,"%d %d\n", nseqs, NPARAMS+1);
   }
   else
   {   
      estfile= fopen("seqs.est", "a");
   }
   fprintf(estfile,"\n%d %d %12.10E\n", dum1,dum2, -mhood);

   for (ktrack1=0;ktrack1 <= NPARAMS; ktrack1++)
   {
      printf("PARAMETER %1d = %12.10E error = ", ktrack1, pvec[ktrack1]);
      fprintf(estfile,"%1d %12.10E ", ktrack1, pvec[ktrack1]);
      if (finmask[ktrack1] >=0)
      {
         tcorr= invmat[finmask[ktrack1]][finmask[ktrack1]];
         if (tcorr > 0.0)
         {
            printf("%12.10E\n", sqrt(tcorr));
            fprintf(estfile, "%12.10E\n", sqrt(tcorr));
         }
         else
         {
            printf("nonsense (Boundary problem ?).\n");
            fprintf(estfile, "-1.0\n");
         }
      }
      else
      {
         printf("not calculated.\n");
         fprintf(estfile, "-2.0\n");
      }
   }
   printf("\n");
   fclose(estfile);
   for (ktrack1=0;ktrack1<=NPARAMS;ktrack1++)
   {
      freedmat(hold[ktrack1],numsamps,seq[0][0]);
   }
   freeivec(finmask,NPARAMS);
   freeimat(holdee,numsamps,seq[0][0]);
   freedmat(holdtg,3,max2(seq[0][0],seq[1][0])+1);
   freedmat(f,NUMLETTERS,NUMLETTERS);
   freedmat(fastf,NUMLETTERS,NUMLETTERS);
   if (nvary>=0)
   { 
      freedmat(invmat,nvary,nvary);
      freedmat(derivmat,nvary,nvary);
      freedvec(pvec,nvary);
      freedvec(lono,nvary);
      freedvec(hino,nvary);
      freedmat(hihi,nvary,nvary);
      freedmat(lolo,nvary,nvary);
      freedmat(hilo,nvary,nvary);
   }
}

double 
dubsum(holdtg,hold, holdee,f,fastf,seq,pars) 
/* the log-likelihood of the two sequences */
double **hold[],**holdtg;
int **holdee;
int **seq;
double **f,**fastf;
double *pars;
{
   double *prvec /* ,r,h,denom */ ;
   double thissum;
   extern double *dvec();
   extern void freedvec();
   double dosum(); /* does the work */
   void prepsum(); /* Prepares for dosum() */

   prvec=dvec(6);
/*
      h=(1.0-r)*(seq[0][0]+seq[1][0]);
      denom= 2.0*(h+2.0*(1.0-r));
      pars[0]=( h-4.0*r+ sqrt(h*h+16.0*r))/denom;
*/
   prepsum(pars,f,fastf,prvec);
   thissum= dosum(holdtg,hold,holdee,f, fastf,seq,prvec);
   freedvec(prvec,6);
   return(thissum);
}

#ifdef PROT

void 
dofrq(seq) /* Read and Set-up the PAM or PAM-like matrix */
int **seq;
{
   extern int frombase();

   /* ++pac */
   extern int *ivec();

   FILE *reveqps,*evalfile,*evecfile,*ievecfile;
   int ktrack1,ktrack2,*order;
   char thisamino;

   reveqps=fopen("reveqps","r");
   ktrack1=0;
   while (ktrack1<NUMLETTERS)
   {
      thisamino=getc(reveqps);
      if (frombase(thisamino) < NUMLETTERS)
      {
         fscanf(reveqps,"%lf",&bfrq[frombase(thisamino)]); 
/*       printf("%2d bfrq= %f\n", ktrack1,bfrq[frombase(thisamino)]); 
*/
         ktrack1++;
      }
   }
   bfrq[NUMLETTERS]=1.0; /* This is the "Unknown" amino acid */
   fclose(reveqps);

   evalfile=fopen("evalfile","r");
   ktrack1=0;
   while (ktrack1<NUMLETTERS)
   {
      thisamino=getc(evalfile);
      if (frombase(thisamino) < NUMLETTERS)
      {
         fscanf(evalfile,"%lf",&lneval[frombase(thisamino)]); 
/*       printf("%2d eval= %f\n", ktrack1,lneval[frombase(thisamino)]); 
*/
         lneval[frombase(thisamino)]= log(lneval[frombase(thisamino)]);
         ktrack1++;
      }
   }
   fclose(evalfile);
   
   order=ivec(NUMLETTERS-1);
   evecfile= fopen("evecfile","r");
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      thisamino= getc(evecfile);
      while (thisamino== ' ')
      {
         thisamino=getc(evecfile);
      }
      order[ktrack1]= frombase(thisamino);
   }
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      for (ktrack2=0;ktrack2 < NUMLETTERS;ktrack2++)
      {
         fscanf(evalfile,"%lf", &evec[order[ktrack1]][order[ktrack2]]);
      }
   }
   fclose(evecfile);

   ievecfile= fopen("ievecfile","r");
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      thisamino= getc(ievecfile);
      while (thisamino== ' ')
      {
         thisamino=getc(ievecfile);
      }
      order[ktrack1]= frombase(thisamino);
   }
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      for (ktrack2=0;ktrack2 < NUMLETTERS;ktrack2++)
      {
         fscanf(ievecfile,"%lf", &ievec[order[ktrack1]][order[ktrack2]]);
      }
   }
   fclose(ievecfile);

   freeivec(order,NUMLETTERS-1);
}

/* calculate Dayhoff replacement matrix and put in ftemp for amount of 
   PAMs equal to grate */
void 
dobd(grate,ftemp)
double grate,**ftemp;
{
   extern char tobase();

   double eval[NUMLETTERS],tempmat[NUMLETTERS][NUMLETTERS];
   
   int ktrack1,ktrack2,ktrack3;

   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {
      eval[ktrack1]=exp(grate*lneval[ktrack1]);

/*      printf("%2d eval[%1c]= %1f\n",ktrack1,tobase(ktrack1),eval[ktrack1]);
*/
   }
   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {
      for (ktrack2=0;ktrack2<NUMLETTERS;ktrack2++)
      {
         tempmat[ktrack1][ktrack2]= evec[ktrack1][ktrack2]*eval[ktrack2];
      }
   }


   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {
      ftemp[ktrack1][ktrack1]=0.0;
      for (ktrack3=0;ktrack3<NUMLETTERS;ktrack3++)
      {
         ftemp[ktrack1][ktrack1] += tempmat[ktrack1][ktrack3]*
         ievec[ktrack3][ktrack1];
      }
      ftemp[ktrack1][ktrack1] *= bfrq[ktrack1]; 
      for (ktrack2=ktrack1+1;ktrack2<NUMLETTERS;ktrack2++)
      {
         ftemp[ktrack1][ktrack2]=0.0;
         for (ktrack3=0;ktrack3<NUMLETTERS;ktrack3++)
         {
            ftemp[ktrack1][ktrack2] += tempmat[ktrack1][ktrack3]*
               ievec[ktrack3][ktrack2];
         }
         ftemp[ktrack1][ktrack2] *= bfrq[ktrack1]; 
         if (ftemp[ktrack1][ktrack2]< SMALLNUM) 
         {
            ftemp[ktrack1][ktrack2]=SMALLNUM;
         }
         ftemp[ktrack2][ktrack1]=ftemp[ktrack1][ktrack2];
      }
      ftemp[ktrack1][NUMLETTERS]=bfrq[ktrack1];
      ftemp[NUMLETTERS][ktrack1]=bfrq[ktrack1];
   }
   ftemp[NUMLETTERS][NUMLETTERS]=1.0;
}

#else
/* calculate nucleotide substitution matrix and put in ftemp for within group
   substitution rate wrate and general substitution rate grate */
void 
dobd(grate,wrate,ftemp)
double grate,wrate, **ftemp;
{
   int ktrack1,ktrack2;
   double expgrate, expwrate;
 
   expgrate= exp(-grate);
   expwrate= exp(-wrate);
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      ftemp[ktrack1][ktrack1]= (1.0-expgrate)*bfrq[ktrack1] +
         expgrate*(1.0-expwrate)*bfrq[ktrack1]/gfrq[group[
         ktrack1]] + expgrate*expwrate;
      ftemp[ktrack1][ktrack1] *= bfrq[ktrack1];
      for (ktrack2=ktrack1+1;ktrack2< NUMLETTERS;ktrack2++)
      {
         if (group[ktrack1]==group[ktrack2])
         {
            ftemp[ktrack1][ktrack2]= (1.0-expgrate)*bfrq[ktrack2] +
               expgrate*(1.0-expwrate)*bfrq[ktrack2]/gfrq[group[ktrack2]];
         }
         else
         {
            ftemp[ktrack1][ktrack2]= (1.0-expgrate)*bfrq[ktrack2];
         }
         ftemp[ktrack1][ktrack2] *= bfrq[ktrack1];
         ftemp[ktrack2][ktrack1]= ftemp[ktrack1][ktrack2];
      }
      ftemp[ktrack1][NUMLETTERS]=bfrq[ktrack1];
      ftemp[NUMLETTERS][ktrack1]=bfrq[ktrack1];
   }
   ftemp[NUMLETTERS][NUMLETTERS]=1.0;
}    

/* count frequency of nucleotides in the two sequences.  These frequencies are
   used as the equilibrium frequencies of the nucelotide types.  This
   is not necessarily going to yield maximum likelihood estimates */
void 
dofrq(seq)
int **seq;
{

   int ktrack1;

   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++) bfrq[ktrack1]=SMALLNUM;

   for (ktrack1=1;ktrack1<=seq[0][0];ktrack1++) bfrq[seq[0][ktrack1]] +=1.0;
   for (ktrack1=1;ktrack1<=seq[1][0];ktrack1++) bfrq[seq[1][ktrack1]] +=1.0;

   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      bfrq[ktrack1]= bfrq[ktrack1]/( seq[0][0] + seq[1][0]+NUMLETTERS*SMALLNUM);   }
   group[0]=0;
   group[1]=0;
   group[2]=1;
   group[3]=1;

   for (ktrack1=0;ktrack1< NUMGROUPS;ktrack1++) gfrq[ktrack1]=0.0;
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      gfrq[group[ktrack1]] += bfrq[ktrack1];
   }
   bfrq[NUMLETTERS]=1.0;
}

#endif

/* transform parameters into some quantities that the routines for 
   calculating alignment likelihoods and full (i.e summed over alignment)
   likelihoods can interpret */
void 
prepsum(pars,f,fastf,prvec)
double *pars,**f,**fastf,*prvec;
{
   void dobd();
   double genp1,beta,mu,lambda,lmurat;
#ifdef PROT
   dobd(pars[1],f);
   dobd(pars[1]*pars[5],fastf);
#else
   dobd(pars[1],pars[7],f);
   dobd(pars[1]*pars[5], pars[7]*pars[5], fastf);
#endif
   lmurat=pars[0];
   mu=pars[2];
   lambda=lmurat*pars[2];
   prvec[0]=pars[3];
   beta= (1.0-exp(lambda-mu))/(mu-lambda*exp(lambda-mu));
   genp1= exp(-mu)*(1.0-lambda*beta)*(1.0-prvec[0])*(lmurat);
   prvec[1]= (1.0-pars[4])*genp1;
   prvec[3]= pars[4]*genp1;
   prvec[2]= lambda*beta*(1.0-prvec[0]);
   prvec[4]= (1.0-prvec[0])*(1.0-prvec[0])*(1.0-exp(-mu)-mu*beta)*
   (1.0-lambda*beta)*lmurat;
   prvec[5]= (1.0-lambda*beta)*(1.0-lmurat);
   prvec[6]= pars[6];
}

/* sum the alignment likelihoods - the workhorse routine.  Also places 
   various columns of the full likelihood matrix into the "hold" parameters */
double 
dosum(holdtg,hold,holdee,f,fastf,seq,prvec)
double **holdtg,**hold[7],*prvec;
double **f,**fastf;
int **seq, **holdee;
{
   extern double *dvec();
   extern int *ivec();
   extern void freedvec();
   extern void freeivec();

   double old1,old2,old4,old5,old7,rold1,rold4,rold7,theanswer;
   double specidterm,spec3term,oldsum,roldsum;
   double *sf1,*sf2,*sf3,*sf4,*sf5,*sf6,*sf7,*sum;
   double *lsf1,*lsf2,*lsf3,*lsf4,*lsf5,*lsf6,*lsf7,*lsum;
   double *htg3,*htg0;
   double tsum,tgap, r,p1term,idterm,fsterm,bdterm,limmterm;
   double fasthold,fhold,pi1hold,pi2hold,magnify,micronum,microfy;
   double *h1,*h2,*h3,*h4,*h5,*h6,*h7;
   double *tg0, *ltg0, tg1,tg2, *tg3, *ltg3;
   int *hee;
   int eemax,s1hold,s2hold,*ee,*lee,eemin,oldee1,oldee2,oldee3;
   int twojumps,s1track,s2track,flag1,flag2;
   int temptrack;
   int s1len,s2len;

   r=prvec[0];
   p1term=prvec[1];
   idterm=prvec[2];
   fsterm=prvec[3];
   bdterm=prvec[4];
   limmterm=prvec[5];

   tgap= prvec[6];

   s1len=seq[0][0];
   s2len=seq[1][0]; 
   lsf1=dvec(s2len);
   lsf2=dvec(s2len);
   lsf3=dvec(s2len);
   lsf4=dvec(s2len);
   lsf5=dvec(s2len);
   lsf6=dvec(s2len);
   lsf7=dvec(s2len);
   ltg0=dvec(s2len);
   ltg3=dvec(s2len);

   lsum=dvec(s2len);
   lee= ivec(s2len);
   magnify= exp(1.0*JUMP); /* magnify and micronum help prevent underflow */
   micronum= exp(-1.0*JUMP);
   *lee= JUMP;
   
   /* set up the (0,0) entry of the matrix */
   *ltg0 = magnify*limmterm*(1.0-tgap)*(1.0-tgap);
   *lsf1=0.0;
   *lsf2=0.0;
   *lsf3=0.0;
   *lsf4=0.0;
   *lsf5=0.0;
   *lsf6= 0.0;
   *lsf7=0.0;

   /* set up the (0,1) entry of the matrix */
   tg0= ltg0+1;
   sf1= lsf1+1;
   sf2=lsf2+1;
   sf3=lsf3+1;
   sf4=lsf4+1;
   sf5=lsf5+1;
   sf6=lsf6+1;
   sf7=lsf7+1;
   ee=lee+1;
   *tg0= 0.5*tgap*bfrq[seq[1][1]]*(*ltg0);
   *sf1=0.0;
   *sf2=0.0;
   *sf3= (*ltg0)*bfrq[seq[1][1]] *idterm;
   *sf4=0.0;
   *sf5=0.0;
   *sf6=0.0;
   *sf7=0.0;

   *lsum= *ltg0;
   *ee= *lee;
   sum=lsum+1;
   *sum= *tg0+ *sf3;
   specidterm= r+ idterm;
   magnify *= magnify;
   microfy= 1.0/magnify;
   twojumps= JUMP+JUMP;
   for (s2track=2;s2track<=s2len;s2track++) /* finish first column of matrix */
   {
      tg0++;
      sf1++;
      sf2++;
      sf3++;
      sf4++;
      sf5++;
      sf6++;
      sf7++;
      sum++;
      ee++;
   
      *tg0= *(tg0-1)*bfrq[seq[1][s2track]]*tgap;
      *sf1=0.0;
      *sf2=0.0;
      *sf3=  bfrq[seq[1][s2track]] * (*(sf3-1)*specidterm+(*(tg0-1)*idterm));
      *ee = *(ee-1);
      *sum= *tg0+*sf3;
      if (*sum < micronum)
      {
         *sf3 *= magnify;
         *tg0 *= magnify;
         *ee += twojumps;
         *sum *= magnify;
      }
      *sf4=0.0;
      *sf5=0.0;
      *sf6=0.0;
      *sf7=0.0;
   }
   /* store the first column */
   ee=lee;
   sf1=lsf1;
   sf2=lsf2;
   sf3=lsf3;
   sf4=lsf4;
   sf5=lsf5;
   sf6=lsf6;
   sf7=lsf7;
   tg0=ltg0;
   htg0= &holdtg[0][0];
   hee= &holdee[0][0];
   h1= &hold[0][0][0];
   h2= &hold[1][0][0];
   h3= &hold[2][0][0];
   h4= &hold[3][0][0];
   h5= &hold[4][0][0];
   h6= &hold[5][0][0];
   h7= &hold[6][0][0];

   for (s2track=0;s2track<=s2len;s2track++)
   {
      *htg0 = *tg0;
      *hee = *ee;
      *h1= *sf1;
      *h2= *sf2;
      *h3= *sf3;
      *h4= *sf4;
      *h5= *sf5;
      *h6= *sf6;
      *h7= *sf7;
      hee++;
      h1++;
      h2++;
      h3++;
      h4++;
      h5++;
      h6++;
      h7++;
      htg0++;
      tg0++;
      ee++;
      sf1++;
      sf2++;
      sf3++;
      sf4++;
      sf5++;
      sf6++;
      sf7++;
   }

   /* loop through remaining columns */
   for (s1track=1;s1track<=s1len;s1track++)
   {
      /* set up the 0th entry in column s1track */
      s1hold=seq[0][s1track];
      pi1hold= bfrq[s1hold];
      if (s1track==1) 
         flag1=0;
      else 
         flag1=1;
      oldee1 = *lee;
      if (flag1) 
      {
         *lsf2 = (tg1*idterm+ *lsf2*specidterm) *pi1hold;
         tg1 *= pi1hold*tgap;
      }
      else 
      {
         *lsf2 = *ltg0*pi1hold* idterm;
         tg1 =   *ltg0*pi1hold*tgap*0.5;
      }
      
      oldsum= *lsum;
      *lsum= *lsf2 + tg1;
      if (*lsum< micronum)
      {
         tg1 *= magnify;
         *lsf2 *= magnify;
         *lsum *= magnify;
         *lee += twojumps;
      }

      *lsf1= 0.0;
      *lsf3= 0.0;
      *lsf4= 0.0;
      *lsf5= 0.0;
      *lsf6= 0.0;
      *lsf7= 0.0;
      ee=lee;
      sum=lsum;
      sf1=lsf1;
      sf2=lsf2;
      sf3=lsf3;
      sf4=lsf4;
      sf5=lsf5;
      sf6=lsf6;
      sf7=lsf7;
      for (s2track=1;s2track<=s2len;s2track++)
      {
         /* finish column s1track */
         sf1++;
         sf2++;
         sf3++;
         sf4++;
         sf5++;
         sf6++;
         sf7++;
         sum++;
         ee++;
         old1= *sf1;
         old2 = *sf2;
         old4= *sf4;
         old5 = *sf5;
         old7= *sf7;
         oldee2= *ee;
         oldee3= *(ee-1);
         roldsum=oldsum;
         oldsum= *sum;
         if (s2track==1) 
            flag2=0;
         else 
            flag2=1;
         s2hold=seq[1][s2track];
         pi2hold= bfrq[s2hold];
         fhold= f[s1hold][s2hold];
         fasthold=fastf[s1hold][s2hold];


         if (flag1 && flag2) 
         {
            *sf1 = fhold*(roldsum*p1term+rold1*r);
            *sf4 = pi2hold*pi1hold*(roldsum*bdterm+rold4*r*r);
            *sf7 = fasthold*(roldsum*fsterm+rold7*r);
         }
         else
         {
            *sf1 = fhold*roldsum * p1term;
            *sf4 = pi2hold *pi1hold*roldsum*bdterm;
            *sf7 = fasthold*roldsum * fsterm;
         }
         if (flag1) 
         {
            *sf2= pi1hold*(oldsum*idterm+old2*r);
            *sf5= pi1hold*r*(old5 + old4);
          }
          else 
          {
             *sf5= 0.0;
             *sf2= pi1hold*oldsum*idterm;
          }
  
          spec3term= *(sum-1)-  *(sf2-1) ;
          if (flag2) 
          {
             *sf3= pi2hold*( spec3term*idterm + *(sf3-1)*r);
             *sf6= pi2hold *r *( *(sf4-1) + *(sf6-1) );
          }
          else 
          {
             *sf3=pi2hold*spec3term*idterm;
             *sf6=0.0;
         }

         if (s2track==s2len)
         {
            if (flag1) 
               tg2 = (0.5*oldsum+ tg2)*tgap*pi1hold;    
            else 
               tg2= oldsum*tgap*pi1hold*0.5;
         }
         if (s1track==s1len)
         {
            if (!flag2) 
            {
               *ltg3=0.0; 
               tg3=ltg3+1; 
               *tg3= 0.5*(*(sum-1))*tgap*pi2hold;
            } 
            else
            {
               tg3++;
               *tg3= (*(tg3-1)+ *(sum-1)*0.5)*tgap*pi2hold;
            }
         }

         if (oldee1 != oldee2 || oldee1 != oldee3)
         {
            eemin = min3(oldee1,oldee2,oldee3);
            eemax = max3(oldee1,oldee2,oldee3);
            if (eemax-eemin > twojumps) 
            {
               printf("trouble in routine dosum() !\n");
            }
            if (oldee1 == eemin)  
            {
               *sf1 *= magnify;
               *sf4 *= magnify;
               *sf7 *= magnify;
            }
            if (oldee2 == eemin)
            {
               *sf2 *= magnify;
               *sf5 *= magnify;
               if (s2track==s2len) 
                 tg2 *=magnify;
            }
            if (oldee3 == eemin)
            {
               *sf3 *= magnify;
               *sf6 *= magnify;
               if (s1track==s1len) 
                  *tg3 *=magnify;
            }
            *ee = eemin + twojumps;
         }  
         *sum= *sf1+ *sf2+ *sf3 + *sf4 + *sf5 + *sf6 + *sf7;
         tsum= *sum;
         if ( s1track==s1len) 
            tsum += *tg3;
         if ( s2track==s2len) 
            tsum += tg2;
         if ( tsum < micronum) /* worry about underflow */
         {
            *ee += twojumps;
            *sum *= magnify;
            *sf1 *= magnify;
            *sf2 *= magnify;
            *sf3 *= magnify;
            *sf4 *= magnify;
            *sf5 *= magnify;
            *sf6 *= magnify;
            *sf7 *= magnify;
            if (s1track==s1len) 
               *tg3 *= magnify;
            if (s2track==s2len) 
               tg2 *= magnify;
         }
         else 
         {
            if ( tsum > magnify) /* worry about overflow */
            {
               *ee -= twojumps;
               *sum *= microfy;
               *sf1 *= microfy;
               *sf2 *= microfy;
               *sf3 *= microfy;
               *sf4 *= microfy;
               *sf5 *= microfy;
               *sf6 *= microfy;
               *sf7 *= microfy;
               if (s1track==s1len) 
                  *tg3 *= microfy;
               if (s2track==s2len) 
                  tg2 *= microfy;
            }
         }
         rold4=old4;
         rold1=old1;
         rold7=old7;
         oldee1=oldee2;
      } 


      if (s1track%SAMPFREQ==0 && s1track != seq[0][0]) 
      { /* then store this column */
         temptrack= s1track/SAMPFREQ;
         holdtg[1][temptrack]=tg1;
         holdtg[2][temptrack]=tg2;
         ee=lee;
         sf1=lsf1;
         sf2=lsf2;
         sf3=lsf3;
         sf4=lsf4;
         sf5=lsf5;
         sf6=lsf6;
         sf7=lsf7;
         hee= &holdee[temptrack][0];
         h1= &hold[0][temptrack][0];
         h2= &hold[1][temptrack][0];
         h3= &hold[2][temptrack][0];
         h4= &hold[3][temptrack][0];
         h5= &hold[4][temptrack][0];
         h6= &hold[5][temptrack][0];
         h7= &hold[6][temptrack][0];


         for (s2track=0;s2track<=s2len;s2track++)
         {
            *hee = *ee;
            *h1= *sf1;
            *h2= *sf2;
            *h3= *sf3;
            *h4= *sf4;
            *h5= *sf5;
            *h6= *sf6;
            *h7= *sf7;
            hee++;
            h1++;
            h2++;
            h3++;
            h4++;
            h5++;
            h6++;
            h7++;
            ee++;
            sf1++;
            sf2++;
            sf3++;
            sf4++;
            sf5++;
            sf6++;
            sf7++;
         }
      }
   }
   theanswer=  *(lee+s2len)-log(tsum);
   htg3= &holdtg[3][0];
   tg3= ltg3;
   for (s2track=0;s2track<=s2len;s2track++)
   {
      *htg3= *tg3;
      htg3++;
      tg3++;
   }
   holdtg[1][0]=0.0;
   holdtg[2][0]=0.0;

   freeivec(lee,s2len);
   freedvec(lsf1,s2len);
   freedvec(lsf2,s2len);
   freedvec(lsf3,s2len);
   freedvec(lsf4,s2len);
   freedvec(lsf5,s2len);
   freedvec(lsf6,s2len);
   freedvec(lsf7,s2len);
   freedvec(lsum,s2len);
   freedvec(ltg0,s2len);
   freedvec(ltg3,s2len);
   return(theanswer);
}

/* fill columns of the alignment likelihood matrix  specified by b1 to e1 */
void 
fillin(f,fastf,seq,b1,e1,sf1,sf2,sf3, sf7,tg0,tg1,tg2,tg3,ee,prvec,rrflag)
int **seq,b1,e1,**ee,rrflag;
double **sf1,**sf2,**sf3,**sf7,*tg0,*tg1,*tg2,*tg3;
double *prvec,**f,**fastf;
{
   extern double *dvec();
   extern void freedvec();

   double fsfac,slfac,sp3maxd,*maxd,oldmaxd,roldmaxd;
   double maxfs,maxid,maxp1,pi1hold,pi2hold;
   double r,microfy,sfmax,magnify,micronum;
   double tgap,p1term,idterm,fsterm,limmterm;
   int seq1len,seq2len;
   int highband,lowband,s1hold,s2hold;
   int bspot2,flag2h,startband, eemax,twojumps;
   int s1place,s1track,s2track;
   
   seq1len=seq[0][0];
   seq2len=seq[1][0];
   maxd=dvec(2*BANDWIDTH+1);
   r=prvec[0];
   p1term=prvec[1];
   idterm=prvec[2];
   fsterm=prvec[3];
   limmterm=prvec[5];
   tgap= prvec[6];
   magnify= exp(1.0*JUMP);
   micronum= exp(-1.0*JUMP);

   if (b1==0) /* then we need to set some things up */
   {
      tg0[BANDWIDTH]=(1.0-tgap)*(1.0-tgap)*limmterm*magnify;
      sf1[0][BANDWIDTH]=0.0;
      sf2[0][BANDWIDTH]= 0.0;
      sf3[0][BANDWIDTH]= 0.0;
      sf7[0][BANDWIDTH]=0.0;
      ee[0][BANDWIDTH] = JUMP;
      tg0[BANDWIDTH+1]= 0.5*tgap*tg0[BANDWIDTH]*bfrq[seq[1][1]]; 
      sf1[0][BANDWIDTH+1]= 0.0;
      sf2[0][BANDWIDTH+1]= 0.0;
      sf3[0][BANDWIDTH+1]= tg0[BANDWIDTH]*bfrq[seq[1][1]]*idterm;
      sf7[0][BANDWIDTH+1]= 0.0;
      ee[0][BANDWIDTH+1]= JUMP;
      maxd[BANDWIDTH]=tg0[BANDWIDTH];
      maxd[BANDWIDTH+1]=max2(tg0[BANDWIDTH+1],sf3[0][BANDWIDTH+1]);
   }

   if ( (rrflag % 2) == 0)
   {
      maxid= max2(idterm,r);
      maxp1 = max2(p1term,r);
      maxfs = max2(fsterm,r);
   }
   else 
   {
      /* this means rrflag %2 == 1 ... */
      maxid= idterm+r;
      maxp1= p1term+r;
      maxfs= fsterm+r;
   }

   highband=min2(2*BANDWIDTH,BANDWIDTH+seq2len);
   magnify *= magnify;
   microfy= 1.0/magnify;
   twojumps= JUMP+ JUMP;
   if (b1 !=0)
   {
      startband=max2(0,BANDWIDTH-b1);
      lowband= max2(0,b1-BANDWIDTH);
      for (s2track=startband; s2track<=highband;s2track++)
      {
         maxd[s2track]= max2(max2(sf1[0][s2track],sf2[0][s2track]), max2(
         sf3[0][s2track], sf7[0][s2track]));
      }
      if (lowband==0) 
         maxd[startband]=max2(maxd[startband],tg1[0]);
   }
   else 
   {
      /* this means b1 == 0 */
      for (s2track=BANDWIDTH+2;s2track<= highband;s2track++)
      {
         tg0[s2track] = tg0[s2track-1]*bfrq[seq[1][s2track-BANDWIDTH]]*tgap;
         sf1[0][s2track]= 0.0;
         sf2[0][s2track]= 0.0;
         sf3[0][s2track]= bfrq[seq[1][s2track-BANDWIDTH]] *
         max2(sf3[0][s2track-1] * maxid, tg0[s2track-1]*idterm);
         sf7[0][s2track] = 0.0;
         ee[0][s2track]=ee[0][s2track-1];
         maxd[s2track]=max2(tg0[s2track],sf3[0][s2track]);
         if (maxd[s2track] < micronum)
         {
            tg0[s2track] *= magnify;
            sf3[0][s2track] *= magnify;
            maxd[s2track] *= magnify;
            ee[0][s2track] += twojumps;
         }
      }
   }
   for (s1track=1;s1track<=e1-b1;s1track++)
   {
      s1place= b1+s1track;
      lowband= max2(0,s1place-BANDWIDTH);
      highband=min2(seq2len,s1place+BANDWIDTH);
      startband= max2(0,BANDWIDTH-s1place); 
      s1hold=seq[0][s1place];
      pi1hold= bfrq[s1hold];      

      oldmaxd= maxd[startband+1];
      roldmaxd=maxd[startband];
      if (lowband==0) 
      {
         ee[s1track][startband] = ee[s1track-1][startband+1];
         if (s1place==1)
         {
            tg1[s1track]=0.5*tgap*pi1hold*tg0[BANDWIDTH];
            sf2[s1track][startband]= pi1hold*tg0[BANDWIDTH]*idterm;
         }
         else
         {
            tg1[s1track]= tg1[s1track-1]*pi1hold*tgap;
            sf2[s1track][startband]= pi1hold*max2(tg1[s1track-1]*idterm,
               sf2[s1track-1][startband+1]*maxid);
         }
         sf1[s1track][startband]= 0.0;
         sf7[s1track][startband]= 0.0;
         if (max2(tg1[s1track],sf2[s1track][startband]) < micronum)
         {
            tg1[s1track] *= magnify;
            sf2[s1track][startband] *= magnify;
            ee[s1track][startband] += twojumps;
         }
      }
      else 
      {
         sf2[s1track][startband]= pi1hold*max2(oldmaxd*idterm,
            sf2[s1track-1][startband+1]*maxid);
         sf1[s1track][startband]= f[s1hold][seq[1][lowband]]*
             max2(sf1[s1track-1][startband]*maxp1, roldmaxd*p1term);
         sf7[s1track][startband]= fastf[s1hold][seq[1][lowband]]*
             max2(sf7[s1track-1][startband]*maxfs, roldmaxd*fsterm);
         if (ee[s1track-1][startband] != ee[s1track-1][startband+1])
         {
            eemax = max2( ee[s1track-1][startband], ee[s1track-1][startband+1]);
            if (eemax> ee[s1track-1][startband]) 
            {
       sf1[s1track][ startband] *= magnify;
       sf7[s1track][startband] *= magnify;
            }
    else 
            {
               sf2[s1track][startband] *= magnify;
            }
            ee[s1track][startband] = eemax;
         }
         else 
            ee[s1track][startband]= ee[s1track-1][startband+1];
      }
      oldmaxd=maxd[startband+1];
      sp3maxd=max2(sf1[s1track][startband], sf7[s1track][startband]);
      if (lowband==0) 
         sp3maxd=max2(sp3maxd,tg1[s1track]);
      maxd[startband]=max2(sp3maxd,sf2[s1track][startband]);
      sfmax=maxd[startband];
      if (sfmax < micronum) 
      {
         ee[s1track][startband] += twojumps;
         sf1[s1track][startband] *=magnify;
         sf2[s1track][startband] *=magnify;
         sf7[s1track][startband] *=magnify;
         maxd[startband] *= magnify;
         sp3maxd *= magnify;
         if (lowband==0) 
            tg1[s1track] *= magnify;
      }

      sf3[s1track][startband]= 0.0;

      for (s2track=lowband+1;s2track<=highband;s2track++)
      {
         bspot2=startband+s2track-lowband;
         roldmaxd=oldmaxd;
         s2hold=seq[1][s2track];
         pi2hold= bfrq[s2hold];
         if (bspot2==2*BANDWIDTH) 
            flag2h=0;
         else 
            flag2h=1;
         if (flag2h) 
            oldmaxd=maxd[bspot2+1];
         else 
            oldmaxd=0.0;
         if (s1place ==1 || s2track==1) 
         {
            slfac= p1term;
            fsfac= fsterm;
         }
         else 
         {
            slfac=maxp1;
            fsfac=maxfs;
         }
         sf7[s1track][bspot2] = fastf[s1hold][s2hold] * max2(
            sf7[s1track-1][bspot2]*fsfac, roldmaxd*fsterm);
         sf1[s1track][bspot2] = f[s1hold][s2hold] * max2(
            sf1[s1track-1][bspot2]*slfac, roldmaxd*p1term);
         if (flag2h)
         {
            sf2[s1track][bspot2]= pi1hold* 
               max2(sf2[s1track-1][bspot2+1] * maxid, oldmaxd*idterm);
         }
         else 
            sf2[s1track][bspot2]= 0.0;
         sf3[s1track][bspot2]= pi2hold*max2( sp3maxd*idterm,
            sf3[s1track][bspot2-1]* maxid );

         if (s1place==seq1len) 
         {
            if (s2track==1)
            {
               tg3[bspot2]= 0.5*tgap*pi2hold*maxd[bspot2-1];
            }
            else 
            {
               tg3[bspot2] = tgap*pi2hold*max2( 
                  tg3[bspot2-1],maxd[bspot2-1]*0.5); 
            }
         }
         if (s2track==seq2len) 
         {
            if (flag2h)
            {
               if (s1place==1)
               {
                 tg2[s1track]=0.5*tgap*pi1hold*oldmaxd;
               }
               else
               {
                  tg2[s1track]=tgap*pi1hold*max2(tg2[s1track-1], 0.5*oldmaxd);
               }
            }   
            else 
               tg2[s1track]=0.0;
         }

         if ((flag2h && ee[s1track-1][bspot2] != ee[s1track-1][bspot2+1])
            || ee[ s1track-1][bspot2] != ee[s1track][bspot2 -1])
         {
            eemax = max3( ee[s1track-1][bspot2],
               flag2h*ee[s1track-1][bspot2+1], ee[s1track][bspot2-1]);
            if (eemax> ee[s1track-1][bspot2]) 
            {
               sf1[s1track][ bspot2] *= magnify;
               sf7[s1track][ bspot2] *= magnify;
            }
            if (flag2h && eemax> ee[s1track-1][bspot2+1]) 
            {
               sf2[s1track][bspot2] *= magnify;
               if (s2track==seq2len) 
                  tg2[s1track] *= magnify;
            }
            if (eemax> ee[s1track][bspot2-1]) 
            {
               sf3[s1track][bspot2] *= magnify;
               if (s1place==seq1len) 
                  tg3[bspot2] *= magnify;
            }
            ee[s1track][bspot2] = eemax;
         }
         else 
            ee[s1track][bspot2] = ee[s1track-1][bspot2];
         sp3maxd= max3(sf1[s1track][bspot2],sf7[s1track][bspot2], 
            sf3[s1track][bspot2]);
         maxd[bspot2]= max2(sf2[s1track][bspot2],sp3maxd);
         sfmax=maxd[bspot2];
         if (s1place==seq1len) 
            sfmax=max2(sfmax,tg3[bspot2]);
         if (s2track==seq2len) 
            sfmax=max2(sfmax,tg2[s1track]);


         if (sfmax < micronum)
         {
            sf1[s1track][bspot2] *= magnify;
            sf2[s1track][bspot2] *= magnify;
            sf3[s1track][bspot2] *= magnify;
            sf7[s1track][bspot2] *= magnify;
            if (s1place==seq1len) 
               tg3[bspot2] *= magnify;
            if (s2track==seq2len) 
               tg2[s1track] *= magnify;
            maxd[bspot2] *= magnify;
            sp3maxd *= magnify;
            ee[s1track][bspot2] += twojumps;
         }
         else 
         {
            if (sfmax > magnify)
            {
               sf1[s1track][bspot2] *= microfy;
               sf2[s1track][bspot2] *= microfy;
               sf3[s1track][bspot2] *= microfy;
               sf7[s1track][bspot2] *= microfy;
               if (s1place==seq1len) 
                  tg3[bspot2] *= microfy;
               if (s2track==seq2len) 
                  tg2[s1track] *=microfy;
               maxd[bspot2] *= microfy;
               sp3maxd *= microfy;
               ee[s1track][bspot2]  -= twojumps;
            }
         }
      }
   } 
   freedvec(maxd,2*BANDWIDTH+1);
}

/* what direction should alignment algorithm take next ? */
int 
decidedir(pdir1,pdir2,pdir3,pdir7,pdir8,pdir9)
double pdir1,pdir2,pdir3,pdir7,pdir8,pdir9;
{
   int nextdir;
   double maxlow,maxhigh;
  
   maxlow=max3(pdir1,pdir2,pdir3);
   maxhigh=max3(pdir7,pdir8,pdir9);

   if (max2(maxlow,maxhigh) ==maxlow)
   {
      if (pdir1==maxlow) 
         nextdir=1;
      else
      {
         if (pdir2==maxlow) 
            nextdir=2;
         else 
            nextdir=3;
      }
   }
   else
   {
      if (pdir7==maxhigh) 
         nextdir=7;
      else
      {
         if (pdir8==maxhigh) 
            nextdir=8;
         else 
            nextdir=9;
      }
   }
   return(nextdir);
}

/* find the alignment and its likelihood */
double 
doali(seq,pars,rflag)
double *pars;
int **seq,rflag;
{
   void fillin();
   int decidedir();
   extern void prepsum();
   extern double **dmat();
   extern int **imat();
   extern double *dvec();
   extern void freedmat();
   extern void freeimat();
   extern void freedvec();

   double pdir8,pdir9,tfac,maxid,maxp1,maxfs,alimax;
   double *htg1,*htg2;
   double **sf1,**sf2,**sf3,**sf7;
   double *tg0,*tg1,*tg2,*tg3;
   double **hsf1,**hsf2,**hsf3,**hsf7;
   double r,*prvec,**f, **fastf, p1term,idterm,fsterm;
   int **hee,s1place,s1hold,eealimax,track2,**ee;
   int dir,s1track,s2track,track;
   int s2place,seq1len,seq2len;

   prvec=dvec(6);
   f=dmat(NUMLETTERS,NUMLETTERS);
   fastf=dmat(NUMLETTERS,NUMLETTERS);
   prepsum(pars,f,fastf,prvec);
   r=prvec[0];
   p1term=prvec[1];
   idterm=prvec[2];
   fsterm=prvec[3];
   seq1len=seq[0][0];
   seq2len=seq[1][0];
   sf1=dmat(SAMPFREQ,2*BANDWIDTH);
   sf2=dmat(SAMPFREQ,2*BANDWIDTH);
   sf3=dmat(SAMPFREQ,2*BANDWIDTH);
   sf7=dmat(SAMPFREQ,2*BANDWIDTH);
   tg0=dvec(2*BANDWIDTH);
   tg1=dvec(SAMPFREQ);
   tg2=dvec(SAMPFREQ);
   tg3=dvec(2*BANDWIDTH);

   ee= imat(SAMPFREQ,2*BANDWIDTH);
   hee =imat((seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   hsf1=dmat((seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   hsf2=dmat((seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   hsf3=dmat((seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   hsf7=dmat((seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   htg1=dvec((seq1len-1)/SAMPFREQ);
   htg2=dvec((seq1len-1)/SAMPFREQ);
   if ( (rflag % 2) == 0) /* should we sum over fragment boundaries ? */
   {
      maxid= max2(idterm,r);
      maxp1 = max2(p1term,r);
      maxfs = max2(fsterm,r);
   }
   else 
   {
      /* this means (rflag % 2) == 1 ... */
      maxid=  idterm+r;
      maxp1 = p1term+r;
      maxfs = fsterm+r;
   }

   /* fill the alignment matrix and store appropriate columns to save
      memory */
   for (track=0;track<= (seq1len-1)/SAMPFREQ;track++)
   {

      fillin(f,fastf,seq,track*SAMPFREQ,min2( seq1len,
      (track+1)*SAMPFREQ), sf1,sf2,sf3,sf7,tg0,tg1,tg2,tg3,ee,prvec,rflag);
      htg1[track]=tg1[0];
      htg2[track]=tg2[0];
      for (s2track=0;s2track<=2*BANDWIDTH;s2track++)
      {

         hsf1[track][s2track]= sf1[0][s2track];
         hsf2[track][s2track]= sf2[0][s2track];
         hsf3[track][s2track]= sf3[0][s2track];
         hsf7[track][s2track]= sf7[0][s2track];
         hee[track][s2track]= ee[0][s2track];

         if (track != (seq1len-1)/SAMPFREQ)
         {
            sf1[0][s2track]= sf1[SAMPFREQ][s2track];
            sf2[0][s2track]= sf2[SAMPFREQ][s2track];
            sf3[0][s2track]= sf3[SAMPFREQ][s2track];
            sf7[0][s2track]= sf7[SAMPFREQ][s2track];
            ee[0][s2track]= ee[SAMPFREQ][s2track];
         }
      }
      if (track != (seq1len-1)/SAMPFREQ)
      {
         tg1[0]=tg1[SAMPFREQ];
         tg2[0]=tg2[SAMPFREQ];
      }
   }

   /* start the backtrack by going to corner of matrix that corresponds to the
      3' end of both sequences */
   s1track=seq1len%SAMPFREQ;
   if (s1track==0 && seq1len>0) 
      s1track=SAMPFREQ;
   s1hold=seq1len-s1track;
   s2track=BANDWIDTH+ seq2len-seq1len;
   eealimax= ee[s1track][s2track]; /* this line & below are alignment 
                                      likelihood */
   alimax=max2(max3(sf1[s1track][s2track], sf2[s1track][s2track], 
          sf3[s1track][ s2track]), max3(sf7[s1track][s2track],tg3[s2track], 
          tg2[s1track]));

   if (rflag < 2)
   {
      alilength=1;
      dir=decidedir(sf1[s1track][s2track], sf2[s1track][s2track], 
         sf3[s1track][s2track],sf7[s1track][s2track],tg3[s2track],tg2[s1track]);

      ali[alilength]=dir;
      s1place=seq1len;
      s2place=seq2len;
      while (s1place > 0 || s2track != BANDWIDTH) 
      /* traceback from matrix corner */
      {
         switch (dir)
         {
            case 1: 
               s1track--;
               if (s1place ==1) 
               {
                  tfac= p1term;
                  pdir8= tg0[s2track]*p1term;
               }
               else 
               {
                  tfac= maxp1;
                  pdir8=0.0;
               }
               if (s2place==1) 
               {
                  pdir9= tg1[s1track]*p1term;
               }
               else
               {
                  pdir9= 0.0;
               }    
               dir=decidedir(sf1[s1track][s2track]*tfac, sf2[s1track][s2track]*
                  p1term, sf3[s1track][s2track]*p1term, sf7[s1track][s2track]*
                  p1term,pdir8,pdir9);
               break;


            case 2: 
               s1track--;
               s2track++;
               if (s1place ==1) 
               {
                  tfac= idterm;
                  pdir8= tg0[s2track]*idterm;
               }
               else 
               {
                  tfac= maxid;
                  pdir8=0.0;
               }
               dir=decidedir(sf1[s1track][s2track]*idterm,sf2[s1track][s2track]*
                  maxid, sf3[s1track][s2track]*idterm, sf7[s1track][s2track]*
                  idterm,pdir8,0.0);
               break;

            case 3:
               s2track--;
               if (s2place==1) 
               {
                  tfac= idterm;
                  pdir9= tg1[s1track]*idterm;
               }
               else
               {
                  tfac=maxid;
                  pdir9= 0.0;
               }    
               dir=decidedir(sf1[s1track][s2track]*idterm, 0.0,
                  sf3[s1track][s2track]*tfac, sf7[s1track][s2track]*
                  idterm,0.0,pdir9);
               break;

            case 7: 
               s1track--;
               if (s1place ==1) 
               {
                  tfac= fsterm;
                  pdir8= tg0[s2track]*fsterm;
               }
               else 
               {
                  tfac= maxfs;
                  pdir8=0.0;
               }
               if (s2place==1) 
               {
                  pdir9= tg1[s1track]*fsterm;
               }
               else
               {
                  pdir9= 0.0;
               }    

               dir=decidedir(sf1[s1track][s2track]*fsterm,sf2[s1track][s2track]*
                  fsterm, sf3[s1track][s2track]*fsterm, sf7[s1track][s2track]*
                  maxfs,pdir8,pdir9);

               break;

            case 8:
               s2track--;
               if (s1place==seq1len) 
               {
                  if (s2place==1)
                  {
                     pdir8=0.0;
                     pdir9=tg1[s1track];
                  }
                  else
                  {
                     pdir8=2.0*tg3[s2track];
                     pdir9=0.0;
                  }
                  dir=decidedir(sf1[s1track][s2track], sf2[s1track][s2track],
                     sf3[s1track][s2track], sf7[s1track][s2track] ,pdir8,pdir9);
               }
               else 
               {
                  /* this means s1place==0 ... */
                  dir=8;
               }

               break;
            case 9:
               s1track--;
               s2track++;
               if (s2place==seq2len)
               {
                  if (s1place==1)
                  {
                     pdir8=tg0[s2track];
                     pdir9=0.0;
                  }
                  else
                  {
                     pdir8=0.0;
                     pdir9=2.0*tg2[s1track];
                  }
                  dir=decidedir(sf1[s1track][s2track], sf2[s1track][s2track],
                     sf3[s1track][s2track], sf7[s1track][s2track] ,pdir8,pdir9);
               }  
               else 
               {
                  /* this means s2place==0 ... */
                  dir=9;
               }  

               break;
         }  
         alilength++;
         ali[alilength]=dir;
         s1place=s1track+s1hold;
         s2place=s1place+s2track-BANDWIDTH;
         /* recalculate next portion of alignment likelihood matrix? */
         if (s1track==0  && (s1hold !=0 || s2track==BANDWIDTH) )
         {
            s1hold -= SAMPFREQ;
            if (s1hold >= 0)
            {
               tg1[0]=htg1[s1hold/SAMPFREQ];
               tg2[0]=htg2[s1hold/SAMPFREQ];
               for (track2=0;track2<=2*BANDWIDTH;track2++)
               {
                  sf1[0][track2]= hsf1[s1hold/SAMPFREQ][track2];
                  sf2[0][track2]= hsf2[s1hold/SAMPFREQ][track2];
                  sf3[0][track2]= hsf3[s1hold/SAMPFREQ][track2];
                  sf7[0][track2]= hsf7[s1hold/SAMPFREQ][track2];
                  ee[0][track2]= hee[s1hold/SAMPFREQ][track2];
               }
               fillin(f,fastf,seq,s1hold,s1hold+SAMPFREQ,sf1,sf2,sf3,sf7,tg0,
                  tg1,tg2,tg3,ee, prvec,rflag);
               s1track=SAMPFREQ;
            }
         }
 
      }
      alilength--;
   }
   freedvec(tg0,2*BANDWIDTH);
   freedvec(tg1,SAMPFREQ);
   freedvec(tg2,SAMPFREQ);
   freedvec(tg3,2*BANDWIDTH);
   freedmat(sf1,SAMPFREQ,2*BANDWIDTH);
   freedmat(sf2,SAMPFREQ,2*BANDWIDTH);
   freedmat(sf3,SAMPFREQ,2*BANDWIDTH);
   freedmat(sf7,SAMPFREQ,2*BANDWIDTH);
   freeimat(ee,SAMPFREQ,2*BANDWIDTH);
   freedvec(prvec,6);
   freedmat(f,NUMLETTERS,NUMLETTERS);
   freedmat(fastf,NUMLETTERS,NUMLETTERS);
   freeimat(hee,(seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   freedmat(hsf1,(seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   freedmat(hsf2,(seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   freedmat(hsf3,(seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   freedmat(hsf7,(seq1len-1)/SAMPFREQ,2*BANDWIDTH);
   freedvec(htg1,(seq1len-1)/SAMPFREQ);
   freedvec(htg2,(seq1len-1)/SAMPFREQ);

   alimax= (double) (eealimax- log(alimax));
   return(alimax);
}

void 
printali(seq) /* print out the alignment */
int **seq;
{
   extern char tobase();
   void botline();

   int numchars,begstep,bseq2,s1track,s2track,alistep;

   numchars=0;
   printf("\n");
   alistep=alilength;
   s1track=1;
   s2track=1;
   bseq2=1;
   begstep=alilength;
   while (alistep>0)
   {
      if (ali[alistep] !=3 && ali[alistep] != 8) 
      {
         printf("%1c",tobase(seq[0][s1track]) );
         s1track++;
         if (ali[alistep]==1 || ali[alistep]==7) 
            s2track++;
      }
      else 
      {
         printf("-");
         s2track++;
      }
      numchars++;
      if (numchars==MLINE) 
      {
         botline(begstep,seq,bseq2);
         begstep=alistep-1;
         bseq2=s2track;
         numchars=0;
      }
      alistep--;
   }
   if (numchars != 0) 
      botline(begstep,seq,bseq2);
}

void 
botline(bstep,seq,bsqn2) 
/* print the bottom line of the alignment */
int bstep,**seq,bsqn2;
{
   extern char tobase();
   int s2track,alistep;
   int numchars;

   s2track=bsqn2;

   numchars=0;
   alistep=bstep;
   printf("\n");
   while (alistep !=0 && numchars != MLINE)
   {
      if (ali[alistep] != 2 && ali[alistep] != 9)
      {
         printf("%1c",tobase(seq[1][s2track]) );
         s2track++;
      }
      else
      {
         printf("-");
      }
      alistep--;
      numchars++;
   }

   numchars=0;
   alistep=bstep;
   printf("\n");
   while (alistep !=0 && numchars != MLINE)
   {
      switch (ali[alistep])
      {
         case 1: printf("s");
            break;
         case 7: printf("f");
            break;
         default: printf(" ");
            break;
      }
      alistep--;
      numchars++;
   }

   printf("\n\n");
}

